# Copyright 2023 Cognite AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import defaultdict
from collections.abc import Callable, Hashable, Iterable, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Literal, final

from cognite.client.data_classes import capabilities as cap
from cognite.client.data_classes.capabilities import (
    Capability,
    GroupsAcl,
    SecurityCategoriesAcl,
)
from rich import print
from rich.console import Console
from rich.markup import escape

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.http_client import ToolkitAPIError
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import SpaceReference
from cognite_toolkit._cdf_tk.client.resource_classes.group import (
    GroupRequest,
    GroupResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import (
    ExternalId,
    InternalId,
    InternalIdUnwrapped,
    NameId,
    RawDatabaseId,
    RawTableId,
)
from cognite_toolkit._cdf_tk.client.resource_classes.securitycategory import (
    SecurityCategoryRequest,
    SecurityCategoryResponse,
)
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.exceptions import ToolkitWrongResourceError
from cognite_toolkit._cdf_tk.resource_classes import GroupYAML, SecurityCategoriesYAML
from cognite_toolkit._cdf_tk.tk_warnings import (
    HighSeverityWarning,
    MediumSeverityWarning,
)
from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils.diff_list import diff_list_hashable, diff_list_identifiable, hash_dict
from cognite_toolkit._cdf_tk.utils.file import sanitize_filename


@dataclass
class _ReplaceMethod:
    """This is a small helper class used in the
    lookup and replace in the ACL scoped ids"""

    lookup_method: Callable[[str, bool], int | None]
    reverse_lookup_method: Callable[[int], str | None]
    id_name: str


class GroupCRUD(ResourceCRUD[NameId, GroupRequest, GroupResponse]):
    folder_name = "auth"
    kind = "Group"
    resource_cls = GroupResponse
    resource_write_cls = GroupRequest
    yaml_cls = GroupYAML
    resource_scopes = frozenset(
        {
            cap.IDScope,
            cap.SpaceIDScope,
            cap.DataSetScope,
            cap.TableScope,
            cap.AssetRootIDScope,
            cap.ExtractionPipelineScope,
            cap.IDScopeLowerCase,
        }
    )
    resource_scope_names = frozenset({scope._scope_name for scope in resource_scopes})  # type: ignore[attr-defined]
    _doc_url = "Groups/operation/createGroups"

    def __init__(
        self,
        client: ToolkitClient,
        build_dir: Path | None,
        console: Console | None,
        target_scopes: Literal[
            "all_scoped_only",
            "resource_scoped_only",
        ] = "all_scoped_only",
    ):
        super().__init__(client, build_dir, console)
        self.target_scopes = target_scopes

    @property
    def display_name(self) -> str:
        return f"groups({self.target_scopes.removesuffix('_only')})"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[GroupRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [
                GroupsAcl.Action.Read,
                GroupsAcl.Action.List,
            ]
            if read_only
            else [
                GroupsAcl.Action.Read,
                GroupsAcl.Action.List,
                GroupsAcl.Action.Create,
                GroupsAcl.Action.Delete,
                GroupsAcl.Action.Update,
            ]
        )

        return GroupsAcl(
            actions,
            GroupsAcl.Scope.All(),
        )

    @classmethod
    def get_id(cls, item: GroupRequest | GroupResponse | dict) -> NameId:
        if isinstance(item, dict):
            return NameId(name=item["name"])
        return NameId(name=item.name)

    @classmethod
    def dump_id(cls, id: NameId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def as_str(cls, id: NameId) -> str:
        return sanitize_filename(id.name)

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        from .classic import AssetCRUD
        from .data_organization import DataSetsCRUD
        from .datamodel import SpaceCRUD
        from .extraction_pipeline import ExtractionPipelineCRUD
        from .location import LocationFilterCRUD
        from .raw import RawDatabaseCRUD, RawTableCRUD
        from .timeseries import TimeSeriesCRUD

        for capability in item.get("capabilities", []):
            for acl, content in capability.items():
                if scope := content.get("scope", {}):
                    if space_ids := scope.get(cap.SpaceIDScope._scope_name, []):
                        if isinstance(space_ids, dict) and "spaceIds" in space_ids:
                            for space_id in space_ids["spaceIds"]:
                                yield SpaceCRUD, SpaceReference(space=space_id)
                    if data_set_ids := scope.get(cap.DataSetScope._scope_name, []):
                        if isinstance(data_set_ids, dict) and "ids" in data_set_ids:
                            for data_set_id in data_set_ids["ids"]:
                                yield DataSetsCRUD, ExternalId(external_id=data_set_id)
                    if table_ids := scope.get(cap.TableScope._scope_name, []):
                        for db_name, tables in table_ids.get("dbsToTables", {}).items():
                            yield RawDatabaseCRUD, RawDatabaseId(name=db_name)
                            if isinstance(tables, list):
                                yield from ((RawTableCRUD, RawTableId(db_name=db_name, name=table)) for table in tables)
                            elif isinstance(tables, dict) and "tables" in tables:
                                for table in tables["tables"]:
                                    yield RawTableCRUD, RawTableId(db_name=db_name, name=table)
                    if extraction_pipeline_ids := scope.get(cap.ExtractionPipelineScope._scope_name, []):
                        if isinstance(extraction_pipeline_ids, dict) and "ids" in extraction_pipeline_ids:
                            for extraction_pipeline_id in extraction_pipeline_ids["ids"]:
                                yield ExtractionPipelineCRUD, ExternalId(external_id=extraction_pipeline_id)
                    if asset_root_ids := scope.get(cap.AssetRootIDScope._scope_name, []):
                        if isinstance(asset_root_ids, dict) and "rootIds" in asset_root_ids:
                            for asset_root_id in asset_root_ids["rootIds"]:
                                yield AssetCRUD, asset_root_id
                    if (ids := scope.get(cap.IDScope._scope_name, [])) or (
                        ids := scope.get(cap.IDScopeLowerCase._scope_name, [])
                    ):
                        loader: type[ResourceCRUD] | None = None
                        if acl == cap.DataSetsAcl._capability_name:
                            loader = DataSetsCRUD
                        elif acl == cap.ExtractionPipelinesAcl._capability_name:
                            loader = ExtractionPipelineCRUD
                        elif acl == cap.TimeSeriesAcl._capability_name:
                            loader = TimeSeriesCRUD
                        elif acl == cap.SecurityCategoriesAcl._capability_name:
                            loader = SecurityCategoryCRUD
                        elif acl == cap.LocationFiltersAcl._capability_name:
                            loader = LocationFilterCRUD
                        if loader is not None and isinstance(ids, dict) and "ids" in ids:
                            for id_ in ids["ids"]:
                                if loader in {TimeSeriesCRUD, LocationFilterCRUD, DataSetsCRUD, ExtractionPipelineCRUD}:
                                    yield loader, ExternalId(external_id=id_)
                                elif loader is SecurityCategoryCRUD:
                                    yield loader, NameId(name=id_)
                                else:
                                    yield loader, id_

    def _substitute_scope_ids(self, group: dict[str, Any], is_dry_run: bool, reverse: bool = False) -> dict[str, Any]:
        replace_method_by_acl = self._create_replace_method_by_acl_and_scope()

        for capability in group.get("capabilities", []):
            for acl, values in capability.items():
                scope = values.get("scope", {})
                if len(scope) != 1:
                    # This will raise an error when the group is loaded.
                    continue
                scope_name, _scope_content = next(iter(scope.items()))

                if (acl, scope_name) in replace_method_by_acl:
                    replace_method = replace_method_by_acl[(acl, scope_name)]
                elif scope_name in replace_method_by_acl:
                    replace_method = replace_method_by_acl[scope_name]
                else:
                    continue
                if ids := scope.get(scope_name, {}).get(replace_method.id_name, []):
                    if reverse:
                        cdf_ids = (
                            replace_method.reverse_lookup_method(int_id) if isinstance(int_id, int) else int_id
                            for int_id in ids
                        )
                        values["scope"][scope_name][replace_method.id_name] = [
                            id_ for id_ in cdf_ids if id_ is not None
                        ]
                    else:
                        values["scope"][scope_name][replace_method.id_name] = [
                            replace_method.lookup_method(ext_id, is_dry_run) if isinstance(ext_id, str) else ext_id
                            for ext_id in ids
                        ]
        return group

    def _create_replace_method_by_acl_and_scope(self) -> dict[tuple[str, str] | str, _ReplaceMethod]:
        source = {
            (cap.DataSetsAcl, cap.DataSetsAcl.Scope.ID): _ReplaceMethod(
                self.client.lookup.data_sets.id,
                self.client.lookup.data_sets.external_id,
                id_name="ids",
            ),
            (cap.ExtractionPipelinesAcl, cap.ExtractionPipelinesAcl.Scope.ID): _ReplaceMethod(
                self.client.lookup.extraction_pipelines.id,
                self.client.lookup.extraction_pipelines.external_id,
                id_name="ids",
            ),
            (cap.LocationFiltersAcl, cap.LocationFiltersAcl.Scope.ID): _ReplaceMethod(
                self.client.lookup.location_filters.id,
                self.client.lookup.location_filters.external_id,
                id_name="ids",
            ),
            (cap.SecurityCategoriesAcl, cap.SecurityCategoriesAcl.Scope.ID): _ReplaceMethod(
                self.client.lookup.security_categories.id,
                self.client.lookup.security_categories.external_id,
                id_name="ids",
            ),
            (cap.TimeSeriesAcl, cap.TimeSeriesAcl.Scope.ID): _ReplaceMethod(
                self.client.lookup.time_series.id,
                self.client.lookup.time_series.external_id,
                id_name="ids",
            ),
            cap.DataSetScope: _ReplaceMethod(
                self.client.lookup.data_sets.id,
                self.client.lookup.data_sets.external_id,
                id_name="ids",
            ),
            cap.ExtractionPipelineScope: _ReplaceMethod(
                self.client.lookup.extraction_pipelines.id,
                self.client.lookup.extraction_pipelines.external_id,
                id_name="ids",
            ),
            cap.AssetRootIDScope: _ReplaceMethod(
                self.client.lookup.assets.id,
                self.client.lookup.assets.external_id,
                id_name="rootIds",
            ),
        }
        # Trick to avoid writing _capability_name and _scope_name for each entry.
        return {
            (key[0]._capability_name, key[1]._scope_name) if isinstance(key, tuple) else key._scope_name: method  # type: ignore[attr-defined]
            for key, method in source.items()
        }

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> GroupRequest:
        is_resource_scoped = any(
            any(scope_name in capability.get(acl, {}).get("scope", {}) for scope_name in self.resource_scope_names)
            for capability in resource.get("capabilities", [])
            for acl in capability
        )

        if self.target_scopes == "all_scoped_only" and is_resource_scoped:
            raise ToolkitWrongResourceError()

        if self.target_scopes == "resource_scoped_only" and not is_resource_scoped:
            raise ToolkitWrongResourceError()

        substituted = self._substitute_scope_ids(resource, is_dry_run)
        return GroupRequest._load(substituted)

    def dump_resource(self, resource: GroupResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump()
        local = local or {}
        if not dumped.get("metadata") and "metadata" not in local:
            dumped.pop("metadata", None)
        if not dumped.get("sourceId") and "sourceId" not in local:
            dumped.pop("sourceId", None)
        # RAWAcls are not returned by the API following the spec.
        # If you have a table scoped RAW ACL the spec, and thus user will input
        # tableScope:
        #   dbsToTables
        #    db1:
        #    - tables1
        #    - tables2
        # While the API will return
        # tableScope:
        #   dbsToTables:
        #   db1:
        #     tables: [tables1, tables2]
        # Note the extra keyword 'tables' in the API response.
        for capability in dumped.get("capabilities", []):
            for acl, content in capability.items():
                if acl != cap.RawAcl._capability_name:
                    continue
                if scope := content.get("scope", {}):
                    if table_scope := scope.get(cap.TableScope._scope_name, {}):
                        db_to_tables = table_scope.get("dbsToTables", {})
                        if not db_to_tables:
                            continue
                        for db_name in list(db_to_tables.keys()):
                            tables = db_to_tables[db_name]
                            if isinstance(tables, dict) and "tables" in tables:
                                db_to_tables[db_name] = tables["tables"]
        # When you dump a CDF Group, all the referenced resources should be available in CDF.
        return self._substitute_scope_ids(dumped, is_dry_run=False, reverse=True)

    def create(self, items: Sequence[GroupRequest]) -> list[GroupResponse]:
        if len(items) == 0:
            return []
        return self._create_with_fallback(items, action="create")

    def update(self, items: Sequence[GroupRequest]) -> list[GroupResponse]:
        # We MUST retrieve all the old groups BEFORE we add the new, if not the new will be deleted
        old_groups = self.client.tool.groups.list(all_groups=True)
        created = self._create_with_fallback(items, action="update")
        created_names = {g.name for g in created}
        to_delete = [group for group in old_groups if group.name in created_names]
        if to_delete:
            self._delete(to_delete, check_own_principal=False)
        return created

    def _create_with_fallback(
        self, items: Sequence[GroupRequest], action: Literal["create", "update"]
    ) -> list[GroupResponse]:
        try:
            return self.client.tool.groups.create(items)
        except ToolkitAPIError as e:
            if not (e.code == 400 and "buffer" in e.message.lower() and len(items) > 1):
                raise e
            # Fallback to create one by one
            created_list: list[GroupResponse] = []
            for item in items:
                try:
                    created = self.client.tool.groups.create([item])
                except ToolkitAPIError as e:
                    HighSeverityWarning(f"Failed to {action} group {item.name}. Error: {escape(str(e))}").print_warning(
                        include_timestamp=True, console=self.console
                    )
                else:
                    created_list.extend(created)
            return created_list

    def retrieve(self, ids: Sequence[NameId]) -> list[GroupResponse]:
        names = {id.name for id in ids}
        remote = self.client.tool.groups.list(all_groups=True)
        return [g for g in remote if g.name in names]

    def delete(self, ids: Sequence[NameId]) -> int:
        return self._delete(self.retrieve(ids), check_own_principal=True)

    def _delete(self, delete_candidates: list[GroupResponse], check_own_principal: bool = True) -> int:
        if check_own_principal:
            print_fun = self.console.print if self.console else print
            try:
                # Let's prevent that we delete groups we belong to
                my_groups = self.client.tool.groups.list()
            except ToolkitAPIError as e:
                print_fun(
                    f"[bold red]ERROR:[/] Failed to retrieve the current service principal's groups. Aborting group deletion.\n{e}"
                )
                return 0
            my_source_ids = {g.source_id for g in my_groups if g.source_id}
        else:
            my_source_ids = set()

        to_delete: list[int] = []
        counts_by_name: dict[str, int] = defaultdict(int)
        for group in delete_candidates:
            if group.source_id in my_source_ids:
                HighSeverityWarning(
                    f"Not deleting group {group.name} with sourceId {group.source_id} as it is used by"
                    f"the current service principal. If you want to delete this group, you must do it manually."
                ).print_warning(console=self.console)
            else:
                to_delete.append(group.id)
                counts_by_name[group.name] += 1
        if duplicates := {name for name, count in counts_by_name.items() if count > 1}:
            MediumSeverityWarning(
                f"The following names are used by multiple groups (all will be deleted): {duplicates}"
            ).print_warning(console=self.console)

        failed_deletes = []
        error_str = ""
        try:
            self.client.tool.groups.delete([InternalId(id=id_) for id_ in to_delete])
        except ToolkitAPIError as e:
            if e.missing:
                # Fallback to delete one by one
                for delete_item_id in to_delete:
                    try:
                        self.client.tool.groups.delete([InternalId(id=delete_item_id)])
                    except ToolkitAPIError as inner_e:
                        if inner_e.missing:
                            # If the error is that the group is already deleted, we can ignore it.
                            continue
                        error_str = str(inner_e)
                        failed_deletes.append(delete_item_id)
            else:
                error_str = str(e)
                failed_deletes.extend(to_delete)
        if failed_deletes:
            MediumSeverityWarning(
                f"Failed to delete groups: {humanize_collection(to_delete)}. "
                "These must be deleted manually in the Fusion UI."
                f"Error: {escape(error_str)}"
            ).print_warning(include_timestamp=True, console=self.console)
        return len(to_delete) - len(failed_deletes)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[GroupResponse]:
        yield from self.client.tool.groups.list(all_groups=True)

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path == ("capabilities",):
            return diff_list_identifiable(local, cdf, get_identifier=hash_dict)
        elif json_path[0] == "capabilities":
            # All sublist inside capabilities are hashable
            return diff_list_hashable(local, cdf)
        elif json_path == ("members",):
            return diff_list_hashable(local, cdf)
        return super().diff_list(local, cdf, json_path)


@final
class GroupAllScopedCRUD(GroupCRUD):
    def __init__(self, client: ToolkitClient, build_dir: Path | None, console: Console | None):
        super().__init__(client, build_dir, console, "all_scoped_only")

    @property
    def display_name(self) -> str:
        return "all-scoped groups"


@final
class SecurityCategoryCRUD(ResourceCRUD[NameId, SecurityCategoryRequest, SecurityCategoryResponse]):
    resource_cls = SecurityCategoryResponse
    resource_write_cls = SecurityCategoryRequest
    kind = "SecurityCategory"
    yaml_cls = SecurityCategoriesYAML
    folder_name = "auth"
    dependencies = frozenset({GroupAllScopedCRUD})
    _doc_url = "Security-categories/operation/createSecurityCategories"
    support_update = False

    @property
    def display_name(self) -> str:
        return "security categories"

    @classmethod
    def get_id(cls, item: SecurityCategoryRequest | SecurityCategoryResponse | dict) -> NameId:
        if isinstance(item, dict):
            return NameId(name=item["name"])
        return NameId(name=item.name)

    @classmethod
    def dump_id(cls, id: NameId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_required_capability(
        cls, items: Sequence[SecurityCategoryRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [
                SecurityCategoriesAcl.Action.List,
                SecurityCategoriesAcl.Action.MemberOf,
            ]
            if read_only
            else [
                SecurityCategoriesAcl.Action.Create,
                SecurityCategoriesAcl.Action.Update,
                SecurityCategoriesAcl.Action.MemberOf,
                SecurityCategoriesAcl.Action.List,
                SecurityCategoriesAcl.Action.Delete,
            ]
        )

        return SecurityCategoriesAcl(
            actions,
            SecurityCategoriesAcl.Scope.All(),
        )

    def create(self, items: Sequence[SecurityCategoryRequest]) -> list[SecurityCategoryResponse]:
        return self.client.tool.security_categories.create(items)

    def retrieve(self, ids: Sequence[NameId]) -> list[SecurityCategoryResponse]:
        names = {id.name for id in ids}
        categories = self.client.tool.security_categories.list(limit=None)
        return [c for c in categories if c.name in names]

    def delete(self, ids: Sequence[NameId]) -> int:
        retrieved = self.retrieve(ids)
        if retrieved:
            self.client.tool.security_categories.delete([InternalIdUnwrapped(id=cat.id) for cat in retrieved])
        return len(retrieved)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[SecurityCategoryResponse]:
        for items in self.client.tool.security_categories.iterate(limit=None):
            yield from items



import sys
from typing import Any, Literal

from pydantic import ConfigDict, Field, JsonValue, field_validator
from pydantic_core.core_schema import ValidationInfo

from cognite_toolkit._cdf_tk.client.resource_classes import streamlit_
from cognite_toolkit._cdf_tk.client.resource_classes.annotation import AnnotationStatus, AnnotationType
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import NodeReference
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId, InternalId
from cognite_toolkit._cdf_tk.client.resource_classes.instance_api import TypedViewReference

from .base import BaseModelRequest

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class Filter(BaseModelRequest): ...


class ClassicFilter(Filter):
    asset_subtree_ids: list[ExternalId | InternalId] | None = None
    data_set_ids: list[ExternalId | InternalId] | None = None

    @classmethod
    def from_asset_subtree_and_data_sets(
        cls,
        asset_subtree_id: str | int | list[str | int] | None = None,
        data_set_id: str | int | list[str | int] | None = None,
    ) -> Self:
        return cls(
            asset_subtree_ids=cls._as_internal_or_external_id_list(asset_subtree_id),
            data_set_ids=cls._as_internal_or_external_id_list(data_set_id),
        )

    @classmethod
    def _as_internal_or_external_id_list(
        cls, id: str | int | list[str | int] | None
    ) -> list[ExternalId | InternalId] | None:
        if id is None:
            return None
        ids = id if isinstance(id, list) else [id]
        return [ExternalId(external_id=item) if isinstance(item, str) else InternalId(id=item) for item in ids]


class StreamlitFilter(ClassicFilter):
    creator: str | None = None

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        body = self.model_dump(mode="json", by_alias=camel_case, exclude_unset=True, exclude={"creator"})
        if self.creator is not None:
            body["metadata"] = {"creator": self.creator}
        body["directoryPrefix"] = streamlit_.STREAMLIT_DIRECTORY
        return body


class TransformationFilter(Filter):
    data_set_ids: list[ExternalId | InternalId] | None = None


class TransformationNotificationFilter(Filter):
    transformation_external_id: str | None = None
    destination: str | None = None


class DataModelingFilter(Filter):
    space: str | None = None
    include_global: bool | None = None


class ContainerFilter(DataModelingFilter):
    used_for: Literal["node", "edge", "record", "all"] | None = None


class ViewFilter(DataModelingFilter):
    include_inherited_properties: bool | None = None
    all_versions: bool | None = None


class DataModelFilter(DataModelingFilter):
    all_versions: bool | None = None


class InstanceFilter(Filter):
    instance_type: Literal["node", "edge"] | None = None
    source: TypedViewReference | None = None
    space: list[str] | None = None
    filter: dict[str, JsonValue] | None = None

    @field_validator("filter")
    def only_filter_or_space(cls, v: dict, info: ValidationInfo) -> dict:
        if v is not None and info.data.get("space") is not None:
            raise ValueError("Cannot specify both filter and space in InstanceFilter")
        return v

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        body: dict[str, Any] = {}
        if self.instance_type is not None:
            body["instanceType"] = self.instance_type
        if self.source is not None:
            body["sources"] = [{"source": self.source.dump()}]
        if self.space is not None:
            instance_type = self.instance_type or "node"
            body["filter"] = {
                "in": {
                    "property": [instance_type, "space"],
                    "values": self.space,
                }
            }
        elif self.filter is not None:
            body["filter"] = self.filter
        return body


class AnnotationFilter(Filter):
    annotated_resource_type: Literal["file", "threedmodel"]
    annotated_resource_ids: list[ExternalId | InternalId]
    annotation_type: AnnotationType | None = None
    created_app: str | None = None
    creating_app_version: str | None = None
    creating_user: str | None = None
    status: AnnotationStatus | None = None


class ThreeDAssetMappingFilter(Filter): ...


class ThreeDAssetMappingAssetIdFilter(ThreeDAssetMappingFilter):
    asset_ids: list[int] = Field(max_length=100)


class ThreeDAssetMappingAssetInstanceIdFilter(ThreeDAssetMappingFilter):
    asset_instance_ids: list[NodeReference] = Field(max_length=100)


class ThreeDAssetMapping3DNodeFilter(ThreeDAssetMappingFilter):
    node_ids: list[int] = Field(max_length=100)


class ThreeDAssetMappingTreeIndexFilter(ThreeDAssetMappingFilter):
    tree_indexes: list[int] = Field(max_length=100)


class SequenceRowFilter(Filter):
    external_id: str
    start: int | None = None
    end: int | None = None
    columns: list[str] | None = None


class SimulatorFilter(Filter):
    simulator_external_ids: list[str] | None = None


class EpochTimestampRange(BaseModelRequest):
    model_config = ConfigDict(protected_namespaces=tuple())
    min_: int | None = Field(default=None, alias="min")
    max_: int | None = Field(default=None, alias="max")


class SimulatorModelFilter(SimulatorFilter):
    external_id_prefix: str | None = None
    data_set_ids: list[int] | None = None


class SimulatorModelRevisionFilter(SimulatorFilter):
    model_external_ids: list[str] | None = None
    all_versions: bool | None = None
    created_time: EpochTimestampRange | None = None
    last_updated_time: EpochTimestampRange | None = None


class SimulatorModelRoutineFilter(SimulatorFilter):
    model_external_ids: list[str] | None = None
    simulator_integration_external_ids: list[str] | None = None


class SimulatorModelRoutineRevisionFilter(SimulatorFilter):
    routine_external_ids: list[str] | None = None
    all_versions: bool | None = None
    model_external_ids: list[str] | None = None
    simulator_integration_external_ids: list[str] | None = None
    created_time: EpochTimestampRange | None = None



from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Hashable, Iterable, Sequence
from typing import Any, ClassVar, Generic

from cognite.client.data_classes import Label, LabelDefinition

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.request_classes.filters import ClassicFilter
from cognite_toolkit._cdf_tk.client.resource_classes.asset import AssetAggregateItem, AssetRequest, AssetResponse
from cognite_toolkit._cdf_tk.client.resource_classes.event import EventRequest, EventResponse
from cognite_toolkit._cdf_tk.client.resource_classes.filemetadata import FileMetadataResponse
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId, InternalId
from cognite_toolkit._cdf_tk.client.resource_classes.timeseries import TimeSeriesRequest, TimeSeriesResponse
from cognite_toolkit._cdf_tk.cruds import (
    AssetCRUD,
    DataSetsCRUD,
    EventCRUD,
    FileMetadataCRUD,
    LabelCRUD,
    TimeSeriesCRUD,
)
from cognite_toolkit._cdf_tk.exceptions import ToolkitMissingResourceError, ToolkitNotImplementedError
from cognite_toolkit._cdf_tk.protocols import T_ResourceRequest, T_ResourceResponse
from cognite_toolkit._cdf_tk.utils.aggregators import (
    AssetAggregator,
    AssetCentricAggregator,
    EventAggregator,
    FileAggregator,
    TimeSeriesAggregator,
)
from cognite_toolkit._cdf_tk.utils.cdf import metadata_key_counts
from cognite_toolkit._cdf_tk.utils.fileio import FileReader, SchemaColumn
from cognite_toolkit._cdf_tk.utils.fileio._readers import TableReader
from cognite_toolkit._cdf_tk.utils.useful_types import (
    AssetCentricType,
    JsonVal,
)
from cognite_toolkit._cdf_tk.utils.useful_types2 import AssetCentricResource

from ._base import (
    ConfigurableStorageIO,
    Page,
    StorageIOConfig,
    TableStorageIO,
    TableUploadableStorageIO,
    UploadItem,
)
from .selectors import AssetCentricSelector, AssetSubtreeSelector, DataSetSelector


class AssetCentricIO(
    Generic[T_ResourceResponse],
    TableStorageIO[AssetCentricSelector, T_ResourceResponse],
    ConfigurableStorageIO[AssetCentricSelector, T_ResourceResponse],
    ABC,
):
    RESOURCE_TYPE: ClassVar[AssetCentricType]
    CHUNK_SIZE = 1000
    BASE_SELECTOR = AssetCentricSelector

    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client)
        self._aggregator = self._get_aggregator()
        self._downloaded_data_sets_by_selector: dict[AssetCentricSelector, set[int]] = defaultdict(set)
        self._downloaded_labels_by_selector: dict[AssetCentricSelector, set[str]] = defaultdict(set)

    @abstractmethod
    def _get_aggregator(self) -> AssetCentricAggregator:
        raise NotImplementedError()

    @abstractmethod
    def retrieve(self, ids: Sequence[int]) -> Sequence[T_ResourceResponse]:
        raise NotImplementedError()

    def count(self, selector: AssetCentricSelector) -> int | None:
        if isinstance(selector, DataSetSelector):
            return self._aggregator.count(data_set_external_id=selector.data_set_external_id)
        elif isinstance(selector, AssetSubtreeSelector):
            return self._aggregator.count(hierarchy=selector.hierarchy)
        return None

    def configurations(self, selector: AssetCentricSelector) -> Iterable[StorageIOConfig]:
        data_set_ids = self._downloaded_data_sets_by_selector[selector]
        if data_set_ids:
            data_set_external_ids = [
                ExternalId(external_id=data_set_external_id)
                for data_set_external_id in self.client.lookup.data_sets.external_id(list(data_set_ids))
            ]
            yield from self._configurations(data_set_external_ids, DataSetsCRUD.create_loader(self.client))

        yield from self._configurations(
            [ExternalId(external_id=label) for label in self._downloaded_labels_by_selector[selector]],
            LabelCRUD.create_loader(self.client),
        )

    def _get_classic_filter(self, selector: AssetCentricSelector) -> ClassicFilter:
        if isinstance(selector, DataSetSelector):
            return ClassicFilter.from_asset_subtree_and_data_sets(data_set_id=selector.data_set_external_id)
        elif isinstance(selector, AssetSubtreeSelector):
            return ClassicFilter.from_asset_subtree_and_data_sets(asset_subtree_id=selector.hierarchy)
        else:
            # This selector is for uploads, not for downloading from CDF.
            raise ToolkitNotImplementedError(f"Selector type {type(selector)} not supported for {type(self).__name__}.")

    def _collect_dependencies(
        self,
        resources: Sequence[AssetResponse]
        | Sequence[FileMetadataResponse]
        | Sequence[TimeSeriesResponse]
        | Sequence[EventResponse],
        selector: AssetCentricSelector,
    ) -> None:
        for resource in resources:
            if resource.data_set_id:
                self._downloaded_data_sets_by_selector[selector].add(resource.data_set_id)
            if isinstance(resource, AssetResponse | FileMetadataResponse):
                for label in resource.labels or []:
                    if isinstance(label, str):
                        self._downloaded_labels_by_selector[selector].add(label)
                    elif isinstance(label, Label | LabelDefinition) and label.external_id:
                        self._downloaded_labels_by_selector[selector].add(label.external_id)
                    elif isinstance(label, dict) and "externalId" in label:
                        self._downloaded_labels_by_selector[selector].add(label["externalId"])

    @classmethod
    def _configurations(
        cls,
        ids: Sequence[Hashable],
        loader: DataSetsCRUD | LabelCRUD,
    ) -> Iterable[StorageIOConfig]:
        if not ids:
            return
        items = loader.retrieve(ids)  # type: ignore[arg-type]
        yield StorageIOConfig(
            kind=loader.kind,
            folder_name=loader.folder_name,
            # We know that the items will be labels for LabelLoader and data sets for DataSetsLoader
            value=[loader.dump_resource(item) for item in items],  # type: ignore[arg-type]
        )

    def _create_identifier(self, internal_id: int) -> str:
        return self.create_internal_identifier(internal_id, self.client.config.project)

    @classmethod
    def create_internal_identifier(cls, internal_id: int, project: str) -> str:
        return f"INTERNAL_ID_project_{project}_{internal_id!s}"

    def _populate_data_set_id_cache(
        self, chunk: Sequence[AssetResponse | FileMetadataResponse | TimeSeriesResponse | EventResponse]
    ) -> None:
        data_set_ids = {item.data_set_id for item in chunk if item.data_set_id is not None}
        self.client.lookup.data_sets.external_id(list(data_set_ids))

    def _populate_security_category_cache(self, chunk: Sequence[FileMetadataResponse | TimeSeriesResponse]) -> None:
        security_category_ids: set[int] = set()
        for item in chunk:
            security_category_ids.update(item.security_categories or [])
        self.client.lookup.security_categories.external_id(list(security_category_ids))

    def _populate_asset_id_cache(self, chunk: Sequence[FileMetadataResponse | EventResponse]) -> None:
        asset_ids: set[int] = set()
        for item in chunk:
            asset_ids.update(item.asset_ids or [])
        self.client.lookup.assets.external_id(list(asset_ids))

    def data_to_row(
        self, data_chunk: Sequence[T_ResourceResponse], selector: AssetCentricSelector | None = None
    ) -> list[dict[str, JsonVal]]:
        rows: list[dict[str, JsonVal]] = []
        for chunk in self.data_to_json_chunk(data_chunk, selector):
            if "metadata" in chunk and isinstance(chunk["metadata"], dict):
                metadata = chunk.pop("metadata")
                # MyPy does understand that metadata is a dict here due to the check above.
                for key, value in metadata.items():  # type: ignore[union-attr]
                    chunk[f"metadata.{key}"] = value
            rows.append(chunk)
        return rows


class UploadableAssetCentricIO(
    Generic[T_ResourceResponse, T_ResourceRequest],
    AssetCentricIO[T_ResourceResponse],
    TableUploadableStorageIO[AssetCentricSelector, T_ResourceResponse, T_ResourceRequest],
    ABC,
):
    def _populate_data_set_external_id_cache(self, chunk: Sequence[dict[str, Any]]) -> None:
        data_set_external_ids: set[str] = set()
        for item in chunk:
            data_set_external_id = item.get("dataSetExternalId")
            if isinstance(data_set_external_id, str):
                data_set_external_ids.add(data_set_external_id)
        self.client.lookup.data_sets.id(list(data_set_external_ids))

    def _populate_asset_external_ids_cache(self, chunk: Sequence[dict[str, Any]]) -> None:
        asset_external_id_set: set[str] = set()
        for item in chunk:
            asset_external_ids = item.get("assetExternalIds")
            if isinstance(asset_external_ids, list):
                for asset_external_id_item in asset_external_ids:
                    if isinstance(asset_external_id_item, str):
                        asset_external_id_set.add(asset_external_id_item)
            asset_external_id = item.get("assetExternalId")
            if isinstance(asset_external_id, str):
                asset_external_id_set.add(asset_external_id)
        self.client.lookup.assets.id(list(asset_external_id_set))

    def _populate_security_category_name_cache(self, chunk: Sequence[dict[str, Any]]) -> None:
        security_category_names: set[str] = set()
        for item in chunk:
            security_category_external_ids_list = item.get("securityCategoryNames")
            if isinstance(security_category_external_ids_list, list):
                for security_category_external_id_item in security_category_external_ids_list:
                    if isinstance(security_category_external_id_item, str):
                        security_category_names.add(security_category_external_id_item)
            security_category_external_id = item.get("securityCategoryNames")
            if isinstance(security_category_external_id, str):
                security_category_names.add(security_category_external_id)
        self.client.lookup.security_categories.id(list(security_category_names))

    def rows_to_data(
        self, rows: list[tuple[str, dict[str, JsonVal]]], selector: AssetCentricSelector | None = None
    ) -> Sequence[UploadItem[T_ResourceRequest]]:
        # We need to populate caches for any external IDs used in the rows before converting to resources.
        # Thus, we override this method instead of row_to_resource, and reuse the json_to_resource method that
        # does the cache population.
        return self.json_chunk_to_data([(source_id, self.row_to_json(row)) for source_id, row in rows])

    @classmethod
    def row_to_json(cls, row: dict[str, JsonVal]) -> dict[str, JsonVal]:
        metadata: dict[str, JsonVal] = {}
        cleaned_row: dict[str, JsonVal] = {}
        for key, value in row.items():
            if key.startswith("metadata."):
                metadata_key = key[len("metadata.") :]
                metadata[metadata_key] = value
            else:
                cleaned_row[key] = value
        if metadata:
            cleaned_row["metadata"] = metadata
        return cleaned_row

    def row_to_resource(
        self, source_id: str, row: dict[str, JsonVal], selector: AssetCentricSelector | None = None
    ) -> T_ResourceRequest:
        raise NotImplementedError(
            f"This method should not be called. {type(self).__name__} overrides rows_to_data instead."
        )


class AssetIO(UploadableAssetCentricIO[AssetResponse, AssetRequest]):
    KIND = "Assets"
    RESOURCE_TYPE = "asset"
    SUPPORTED_DOWNLOAD_FORMATS = frozenset({".parquet", ".csv", ".ndjson"})
    SUPPORTED_COMPRESSIONS = frozenset({".gz"})
    SUPPORTED_READ_FORMATS = frozenset({".parquet", ".csv", ".ndjson", ".yaml", ".yml"})
    UPLOAD_ENDPOINT = "/assets"

    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client)
        self._crud = AssetCRUD.create_loader(self.client)

    def as_id(self, item: AssetResponse) -> str:
        return item.external_id if item.external_id is not None else self._create_identifier(item.id)

    def _get_aggregator(self) -> AssetCentricAggregator:
        return AssetAggregator(self.client)

    def get_schema(self, selector: AssetCentricSelector) -> list[SchemaColumn]:
        data_set_ids: list[int] = []
        if isinstance(selector, DataSetSelector):
            data_set_id = self.client.lookup.data_sets.id(selector.data_set_external_id)
            if data_set_id is None:
                raise ToolkitMissingResourceError(
                    f"Data set with external ID {selector.data_set_external_id} not found."
                )
            data_set_ids.append(data_set_id)
        hierarchy: list[int] = []
        if isinstance(selector, AssetSubtreeSelector):
            asset_id = self.client.lookup.assets.id(selector.hierarchy)
            if asset_id is None:
                raise ToolkitMissingResourceError(f"Asset with external ID {selector.hierarchy} not found.")
            hierarchy.append(asset_id)

        if hierarchy or data_set_ids:
            metadata_keys = metadata_key_counts(
                self.client, "assets", data_sets=data_set_ids or None, hierarchies=hierarchy or None
            )
        else:
            metadata_keys = []
        metadata_schema: list[SchemaColumn] = []
        if metadata_keys:
            metadata_schema.extend(
                [SchemaColumn(name=f"metadata.{key}", type="string", is_array=False) for key, _ in metadata_keys]
            )
        asset_schema = [
            SchemaColumn(name="externalId", type="string"),
            SchemaColumn(name="name", type="string"),
            SchemaColumn(name="parentExternalId", type="string"),
            SchemaColumn(name="description", type="string"),
            SchemaColumn(name="dataSetExternalId", type="string"),
            SchemaColumn(name="source", type="string"),
            SchemaColumn(name="labels", type="string", is_array=True),
            SchemaColumn(name="geoLocation", type="json"),
            SchemaColumn(name="childCount", type="integer"),
            SchemaColumn(name="depth", type="integer"),
            SchemaColumn(name="path", type="string", is_array=True),
        ]
        return asset_schema + metadata_schema

    def stream_data(self, selector: AssetCentricSelector, limit: int | None = None) -> Iterable[Page]:
        filter_ = self._get_classic_filter(selector)
        cursor: str | None = None
        total_count = 0
        while True:
            page = self.client.tool.assets.paginate(
                aggregated_properties=True,
                filter=filter_,
                limit=self.CHUNK_SIZE,
                cursor=cursor,
            )
            self._collect_dependencies(page.items, selector)
            yield Page(worker_id="main", items=page.items)
            total_count += len(page.items)
            if page.next_cursor is None or (limit is not None and total_count >= limit):
                break
            cursor = page.next_cursor

    def data_to_json_chunk(
        self, data_chunk: Sequence[AssetResponse], selector: AssetCentricSelector | None = None
    ) -> list[dict[str, JsonVal]]:
        # Ensure data sets are looked up to populate cache.
        # This is to avoid looking up each data set id individually in the .dump_resource call.
        self._populate_data_set_id_cache(data_chunk)
        asset_ids = {
            segment["id"]
            for item in data_chunk
            if isinstance(item.aggregates, AssetAggregateItem)
            for segment in item.aggregates.path
        }
        self.client.lookup.assets.external_id(list(asset_ids))

        return [self._crud.dump_resource(item) for item in data_chunk]

    def json_to_resource(self, item_json: dict[str, JsonVal]) -> AssetRequest:
        return self._crud.load_resource(item_json)

    def retrieve(self, ids: Sequence[int]) -> list[AssetResponse]:
        return self.client.tool.assets.retrieve(InternalId.from_ids(ids))

    @classmethod
    def read_chunks(
        cls, reader: FileReader, selector: AssetCentricSelector
    ) -> Iterable[list[tuple[str, dict[str, JsonVal]]]]:
        """Assets require special handling when reading data to ensure parent assets are created first."""
        current_depth = max_depth = 0
        data_name = "row" if isinstance(reader, TableReader) else "line"
        batch: list[tuple[str, dict[str, JsonVal]]] = []
        # We read the file multiple times, once for each depth level, to ensure parents are created before children.
        while current_depth <= max_depth:
            for line_number, item in reader.read_chunks_with_line_numbers():
                try:
                    depth = int(item["depth"])  # type: ignore[arg-type]
                except (TypeError, ValueError, KeyError):
                    if current_depth == 0:
                        # If depth is not set, we yield it at depth 0
                        batch.append((f"{data_name} {line_number}", item))
                else:
                    if depth == current_depth:
                        batch.append((f"{data_name} {line_number}", item))
                    elif current_depth == 0:
                        max_depth = max(max_depth, depth)
                if len(batch) >= cls.CHUNK_SIZE:
                    yield batch
                    batch = []
            if batch:
                yield batch
                batch = []
            current_depth += 1


class FileMetadataIO(AssetCentricIO[FileMetadataResponse]):
    KIND = "FileMetadata"
    RESOURCE_TYPE = "file"
    SUPPORTED_DOWNLOAD_FORMATS = frozenset({".parquet", ".csv", ".ndjson"})
    SUPPORTED_COMPRESSIONS = frozenset({".gz"})
    SUPPORTED_READ_FORMATS = frozenset({".parquet", ".csv", ".ndjson"})
    UPLOAD_ENDPOINT = "/files"

    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client)
        self._crud = FileMetadataCRUD.create_loader(self.client)

    def as_id(self, item: FileMetadataResponse) -> str:
        return item.external_id if item.external_id is not None else self._create_identifier(item.id)

    def _get_aggregator(self) -> AssetCentricAggregator:
        return FileAggregator(self.client)

    def get_schema(self, selector: AssetCentricSelector) -> list[SchemaColumn]:
        data_set_ids: list[int] = []
        if isinstance(selector, DataSetSelector):
            data_set_id = self.client.lookup.data_sets.id(selector.data_set_external_id)
            if data_set_id is None:
                raise ToolkitMissingResourceError(
                    f"Data set with external ID {selector.data_set_external_id} not found."
                )
            data_set_ids.append(data_set_id)
        if isinstance(selector, AssetSubtreeSelector):
            raise ToolkitNotImplementedError(f"Selector type {type(selector)} not supported for FileIO.")

        if data_set_ids:
            metadata_keys = metadata_key_counts(self.client, "files", data_sets=data_set_ids or None, hierarchies=None)
        else:
            metadata_keys = []
        metadata_schema: list[SchemaColumn] = []
        if metadata_keys:
            metadata_schema.extend(
                [SchemaColumn(name=f"metadata.{key}", type="string", is_array=False) for key, _ in metadata_keys]
            )
        file_schema = [
            SchemaColumn(name="externalId", type="string"),
            SchemaColumn(name="name", type="string"),
            SchemaColumn(name="directory", type="string"),
            SchemaColumn(name="mimeType", type="string"),
            SchemaColumn(name="dataSetExternalId", type="string"),
            SchemaColumn(name="assetExternalIds", type="string", is_array=True),
            SchemaColumn(name="source", type="string"),
            SchemaColumn(name="sourceCreatedTime", type="integer"),
            SchemaColumn(name="sourceModifiedTime", type="integer"),
            SchemaColumn(name="securityCategories", type="string", is_array=True),
            SchemaColumn(name="labels", type="string", is_array=True),
            SchemaColumn(name="geoLocation", type="json"),
        ]
        return file_schema + metadata_schema

    def stream_data(
        self, selector: AssetCentricSelector, limit: int | None = None
    ) -> Iterable[Page[FileMetadataResponse]]:
        filter_ = self._get_classic_filter(selector)
        cursor: str | None = None
        total_count = 0
        while True:
            page = self.client.tool.filemetadata.paginate(
                filter=filter_,
                limit=self.CHUNK_SIZE,
                cursor=cursor,
            )
            self._collect_dependencies(page.items, selector)
            yield Page(worker_id="main", items=page.items)
            total_count += len(page.items)
            if page.next_cursor is None or (limit is not None and total_count >= limit):
                break
            cursor = page.next_cursor

    def retrieve(self, ids: Sequence[int]) -> list[FileMetadataResponse]:
        return self.client.tool.filemetadata.retrieve(InternalId.from_ids(ids))

    def data_to_json_chunk(
        self, data_chunk: Sequence[FileMetadataResponse], selector: AssetCentricSelector | None = None
    ) -> list[dict[str, JsonVal]]:
        # Ensure data sets/assets/security-categories are looked up to populate cache.
        # This is to avoid looking up each data set id individually in the .dump_resource call
        self._populate_data_set_id_cache(data_chunk)
        self._populate_asset_id_cache(data_chunk)
        self._populate_security_category_cache(data_chunk)

        return [self._crud.dump_resource(item) for item in data_chunk]


class TimeSeriesIO(UploadableAssetCentricIO[TimeSeriesResponse, TimeSeriesRequest]):
    KIND = "TimeSeries"
    SUPPORTED_DOWNLOAD_FORMATS = frozenset({".parquet", ".csv", ".ndjson"})
    SUPPORTED_COMPRESSIONS = frozenset({".gz"})
    SUPPORTED_READ_FORMATS = frozenset({".parquet", ".csv", ".ndjson"})
    UPLOAD_ENDPOINT = "/timeseries"
    RESOURCE_TYPE = "timeseries"

    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client)
        self._crud = TimeSeriesCRUD.create_loader(self.client)

    def as_id(self, item: TimeSeriesResponse) -> str:
        return item.external_id if item.external_id is not None else self._create_identifier(item.id)

    def _get_aggregator(self) -> AssetCentricAggregator:
        return TimeSeriesAggregator(self.client)

    def retrieve(self, ids: Sequence[int]) -> list[TimeSeriesResponse]:
        return self.client.tool.timeseries.retrieve(InternalId.from_ids(ids))

    def stream_data(self, selector: AssetCentricSelector, limit: int | None = None) -> Iterable[Page]:
        filter_ = self._get_classic_filter(selector)
        cursor: str | None = None
        total_count = 0
        while True:
            page = self.client.tool.timeseries.paginate(
                filter=filter_,
                limit=self.CHUNK_SIZE,
                cursor=cursor,
            )
            self._collect_dependencies(page.items, selector)
            yield Page(worker_id="main", items=page.items)
            total_count += len(page.items)
            if page.next_cursor is None or (limit is not None and total_count >= limit):
                break
            cursor = page.next_cursor

    def data_to_json_chunk(
        self, data_chunk: Sequence[TimeSeriesResponse], selector: AssetCentricSelector | None = None
    ) -> list[dict[str, JsonVal]]:
        # Ensure data sets/assets/security categories are looked up to populate cache.
        self._populate_data_set_id_cache(data_chunk)
        self._populate_security_category_cache(data_chunk)
        asset_ids = {item.asset_id for item in data_chunk if item.asset_id is not None}
        self.client.lookup.assets.external_id(list(asset_ids))

        return [self._crud.dump_resource(item) for item in data_chunk]

    def json_chunk_to_data(
        self, data_chunk: list[tuple[str, dict[str, JsonVal]]]
    ) -> Sequence[UploadItem[TimeSeriesRequest]]:
        chunks = [item_json for _, item_json in data_chunk]
        self._populate_asset_external_ids_cache(chunks)
        self._populate_data_set_external_id_cache(chunks)
        self._populate_security_category_name_cache(chunks)
        return super().json_chunk_to_data(data_chunk)

    def json_to_resource(self, item_json: dict[str, JsonVal]) -> TimeSeriesRequest:
        return self._crud.load_resource(item_json)

    def get_schema(self, selector: AssetCentricSelector) -> list[SchemaColumn]:
        data_set_ids: list[int] = []
        if isinstance(selector, DataSetSelector):
            data_set_id = self.client.lookup.data_sets.id(selector.data_set_external_id)
            if data_set_id is None:
                raise ToolkitMissingResourceError(
                    f"Data set with external ID {selector.data_set_external_id} not found."
                )
            data_set_ids.append(data_set_id)
        elif isinstance(selector, AssetSubtreeSelector):
            raise ToolkitNotImplementedError(f"Selector type {type(selector)} not supported for {type(self).__name__}.")

        if data_set_ids:
            metadata_keys = metadata_key_counts(
                self.client, "timeseries", data_sets=data_set_ids or None, hierarchies=None
            )
        else:
            metadata_keys = []
        metadata_schema: list[SchemaColumn] = []
        if metadata_keys:
            metadata_schema.extend(
                [SchemaColumn(name=f"metadata.{key}", type="string", is_array=False) for key, _ in metadata_keys]
            )
        ts_schema = [
            SchemaColumn(name="externalId", type="string"),
            SchemaColumn(name="name", type="string"),
            SchemaColumn(name="isString", type="boolean"),
            SchemaColumn(name="unit", type="string"),
            SchemaColumn(name="unitExternalId", type="string"),
            SchemaColumn(name="assetExternalId", type="string"),
            SchemaColumn(name="isStep", type="boolean"),
            SchemaColumn(name="description", type="string"),
            SchemaColumn(name="securityCategories", type="string", is_array=True),
            SchemaColumn(name="dataSetExternalId", type="string"),
        ]
        return ts_schema + metadata_schema


class EventIO(UploadableAssetCentricIO[EventResponse, EventRequest]):
    KIND = "Events"
    SUPPORTED_DOWNLOAD_FORMATS = frozenset({".parquet", ".csv", ".ndjson"})
    SUPPORTED_COMPRESSIONS = frozenset({".gz"})
    SUPPORTED_READ_FORMATS = frozenset({".parquet", ".csv", ".ndjson"})
    UPLOAD_ENDPOINT = "/events"
    RESOURCE_TYPE = "event"

    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client)
        self._crud = EventCRUD.create_loader(self.client)

    def as_id(self, item: EventResponse) -> str:
        return item.external_id if item.external_id is not None else self._create_identifier(item.id)

    def _get_aggregator(self) -> AssetCentricAggregator:
        return EventAggregator(self.client)

    def get_schema(self, selector: AssetCentricSelector) -> list[SchemaColumn]:
        data_set_ids: list[int] = []
        if isinstance(selector, DataSetSelector):
            data_set_id = self.client.lookup.data_sets.id(selector.data_set_external_id)
            if data_set_id is None:
                raise ToolkitMissingResourceError(
                    f"Data set with external ID {selector.data_set_external_id} not found."
                )
            data_set_ids.append(data_set_id)
        hierarchy: list[int] = []
        if isinstance(selector, AssetSubtreeSelector):
            raise ToolkitNotImplementedError(f"Selector type {type(selector)} not supported for {type(self).__name__}.")

        if hierarchy or data_set_ids:
            metadata_keys = metadata_key_counts(
                self.client, "events", data_sets=data_set_ids or None, hierarchies=hierarchy or None
            )
        else:
            metadata_keys = []
        metadata_schema: list[SchemaColumn] = []
        if metadata_keys:
            metadata_schema.extend(
                [SchemaColumn(name=f"metadata.{key}", type="string", is_array=False) for key, _ in metadata_keys]
            )
        event_schema = [
            SchemaColumn(name="externalId", type="string"),
            SchemaColumn(name="dataSetExternalId", type="string"),
            SchemaColumn(name="startTime", type="integer"),
            SchemaColumn(name="endTime", type="integer"),
            SchemaColumn(name="type", type="string"),
            SchemaColumn(name="subtype", type="string"),
            SchemaColumn(name="description", type="string"),
            SchemaColumn(name="assetExternalIds", type="string", is_array=True),
            SchemaColumn(name="source", type="string"),
        ]
        return event_schema + metadata_schema

    def stream_data(self, selector: AssetCentricSelector, limit: int | None = None) -> Iterable[Page]:
        filter_ = self._get_classic_filter(selector)
        cursor: str | None = None
        total_count = 0
        while True:
            page = self.client.tool.events.paginate(
                filter=filter_,
                limit=self.CHUNK_SIZE,
                cursor=cursor,
            )
            self._collect_dependencies(page.items, selector)
            yield Page(worker_id="main", items=page.items)
            total_count += len(page.items)
            if page.next_cursor is None or (limit is not None and total_count >= limit):
                break
            cursor = page.next_cursor

    def data_to_json_chunk(
        self, data_chunk: Sequence[EventResponse], selector: AssetCentricSelector | None = None
    ) -> list[dict[str, JsonVal]]:
        # Ensure data sets/assets are looked up to populate cache.
        self._populate_data_set_id_cache(data_chunk)
        self._populate_asset_id_cache(data_chunk)

        return [self._crud.dump_resource(item) for item in data_chunk]

    def json_chunk_to_data(
        self, data_chunk: list[tuple[str, dict[str, JsonVal]]]
    ) -> Sequence[UploadItem[EventRequest]]:
        chunks = [item_json for _, item_json in data_chunk]
        self._populate_asset_external_ids_cache(chunks)
        self._populate_data_set_external_id_cache(chunks)
        return super().json_chunk_to_data(data_chunk)

    def json_to_resource(self, item_json: dict[str, JsonVal]) -> EventRequest:
        return self._crud.load_resource(item_json)

    def retrieve(self, ids: Sequence[int]) -> list[EventResponse]:
        return self.client.tool.events.retrieve(InternalId.from_ids(ids))


class HierarchyIO(ConfigurableStorageIO[AssetCentricSelector, AssetCentricResource]):
    CHUNK_SIZE = 1000
    BASE_SELECTOR = AssetCentricSelector
    SUPPORTED_DOWNLOAD_FORMATS = frozenset({".parquet", ".csv", ".ndjson"})
    SUPPORTED_COMPRESSIONS = frozenset({".gz"})

    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client)
        self._asset_io = AssetIO(client)
        self._file_io = FileMetadataIO(client)
        self._timeseries_io = TimeSeriesIO(client)
        self._event_io = EventIO(client)
        self._io_by_kind: dict[str, AssetCentricIO] = {
            self._asset_io.KIND: self._asset_io,
            self._file_io.KIND: self._file_io,
            self._timeseries_io.KIND: self._timeseries_io,
            self._event_io.KIND: self._event_io,
        }

    def as_id(self, item: AssetCentricResource) -> str:
        return item.external_id or AssetCentricIO.create_internal_identifier(item.id, self.client.config.project)

    def stream_data(
        self, selector: AssetCentricSelector, limit: int | None = None
    ) -> Iterable[Page[AssetCentricResource]]:
        yield from self.get_resource_io(selector.kind).stream_data(selector, limit)

    def count(self, selector: AssetCentricSelector) -> int | None:
        return self.get_resource_io(selector.kind).count(selector)

    def data_to_json_chunk(
        self, data_chunk: Sequence[AssetCentricResource], selector: AssetCentricSelector | None = None
    ) -> list[dict[str, JsonVal]]:
        if selector is None:
            raise ValueError(f"Selector must be provided to convert data to JSON chunk for {type(self).__name__}.)")
        return self.get_resource_io(selector.kind).data_to_json_chunk(data_chunk, selector)

    def configurations(self, selector: AssetCentricSelector) -> Iterable[StorageIOConfig]:
        yield from self.get_resource_io(selector.kind).configurations(selector)

    def get_resource_io(self, kind: str) -> AssetCentricIO:
        return self._io_by_kind[kind]



from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    RequestResource,
    ResponseResource,
)

from .identifiers import ExternalId


class Workflow(BaseModelObject):
    external_id: str
    description: str | None = None
    data_set_id: int | None = None
    max_concurrent_executions: int | None = None

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)


class WorkflowRequest(Workflow, RequestResource): ...


class WorkflowResponse(Workflow, ResponseResource[WorkflowRequest]):
    created_time: int
    last_updated_time: int

    @classmethod
    def request_cls(cls) -> type[WorkflowRequest]:
        return WorkflowRequest



import contextlib
from pathlib import Path
from typing import Annotated, Any

import typer
from rich import print

from cognite_toolkit._cdf_tk.cdf_toml import CDFToml
from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands import ResourcesCommand
from cognite_toolkit._cdf_tk.feature_flags import FeatureFlag, Flags
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables

from ._run import RunApp

CDF_TOML = CDFToml.load(Path.cwd())


class DevApp(typer.Typer):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.main)
        self.add_typer(RunApp(*args, **kwargs), name="run")
        if FeatureFlag.is_enabled(Flags.CREATE):
            self.command("create")(self.create)

    @staticmethod
    def main(ctx: typer.Context) -> None:
        """Commands to work with development."""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf dev --help[/] for more information.")
        return None

    def create(
        self,
        kind: Annotated[
            list[str] | None,
            typer.Argument(
                help="The kind of resource to create. eg. container, space, view, datamodel, etc.",
                callback=lambda ctx, param, value: [
                    s.strip() for item in value or [] for s in item.split(",") if s.strip()
                ],
            ),
        ] = None,
        module: Annotated[
            str | None,
            typer.Option(
                "--module",
                "-m",
                help="Name of an existing module or a new module to create the resource in.",
            ),
        ] = None,
        prefix: Annotated[
            str | None,
            typer.Option(
                "--prefix",
                "-p",
                help="The prefix of the resource file to create without suffixes and extensions. "
                "eg. --prefix=my_space. If not provided, a default prefix like 'my_<kind>' will be used.",
            ),
        ] = None,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
        organization_dir: Annotated[
            Path,
            typer.Option(
                "--organization-dir",
                "-o",
                help="Path to the organization directory",
            ),
        ] = CDF_TOML.cdf.default_organization_dir,
    ) -> None:
        """create resource YAMLs."""
        client: ToolkitClient | None = None
        with contextlib.redirect_stdout(None), contextlib.suppress(Exception):
            # Try to load client if possible, but ignore errors.
            # This is only used for logging purposes in the command.
            client = EnvironmentVariables.create_from_environment().get_client()

        cmd = ResourcesCommand(client=client)
        cmd.run(
            lambda: cmd.create(
                organization_dir=organization_dir,
                module_name=module,
                kind=kind,
                prefix=prefix,
                verbose=verbose,
            )
        )



from pydantic import Field

from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import ViewReferenceNoVersion

from .base import BaseModelResource, ToolkitResource


class ViewId(BaseModelResource):
    space: str
    external_id: str


class PropertyConfig(BaseModelResource):
    property_name: str = Field(alias="property")
    disabled: bool | None = None
    selected: bool | None = None
    hidden: bool | None = None


class SearchConfigYAML(ToolkitResource):
    view: ViewId
    use_as_name: str | None = None
    use_as_description: str | None = None
    columns_layout: list[PropertyConfig] | None = None
    filter_layout: list[PropertyConfig] | None = None
    properties_layout: list[PropertyConfig] | None = None

    def as_id(self) -> ViewReferenceNoVersion:
        return ViewReferenceNoVersion(space=self.view.space, external_id=self.view.external_id)



"""Groups API for managing CDF access groups.

Based on the API specification at:
https://api-docs.cognite.com/20230101/tag/Groups/operation/createGroups
"""

from collections.abc import Sequence

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, Endpoint, PagedResponse
from cognite_toolkit._cdf_tk.client.http_client import (
    HTTPClient,
    ItemsSuccessResponse,
    RequestMessage,
    SuccessResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.group import GroupRequest, GroupResponse
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import InternalId
from cognite_toolkit._cdf_tk.utils.collection import chunker_sequence


class GroupsAPI(CDFResourceAPI[GroupResponse]):
    """API for managing CDF access groups.

    Note: Groups do not support update operations.
    """

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/groups", item_limit=1000),
                "delete": Endpoint(method="POST", path="/groups/delete", item_limit=1000),
                "list": Endpoint(method="GET", path="/groups", item_limit=1000),
            },
        )

    def _validate_page_response(self, response: SuccessResponse | ItemsSuccessResponse) -> PagedResponse[GroupResponse]:
        return PagedResponse[GroupResponse].model_validate_json(response.body)

    def create(self, items: Sequence[GroupRequest]) -> list[GroupResponse]:
        """Create groups in CDF.

        Args:
            items: List of GroupRequest objects to create.

        Returns:
            List of created GroupResponse objects.
        """
        return self._request_item_response(items, "create")

    def delete(self, items: Sequence[InternalId]) -> None:
        """Delete groups from CDF.

        Args:
            items: List of InternalId objects to delete.
        """
        # Custom implementation since delete does not wrap the items in a {"id": ...} structure
        endpoint = self._method_endpoint_map["delete"]
        for chunk in chunker_sequence(items, endpoint.item_limit):
            request = RequestMessage(
                endpoint_url=self._make_url(endpoint.path),
                method=endpoint.method,
                body_content={"items": [item.id for item in chunk]},
            )
            response = self._http_client.request_single_retries(request)
            response.get_success_or_raise()

    def list(self, all_groups: bool = False) -> list[GroupResponse]:
        """List all groups in CDF.

        Args:
            all_groups: Whether to return all groups (requires admin permissions).

        Returns:
            List of GroupResponse objects.
        """
        endpoint = self._method_endpoint_map["list"]
        response = self._http_client.request_single_retries(
            RequestMessage(
                endpoint_url=self._make_url(endpoint.path),
                method=endpoint.method,
                parameters={"all": all_groups},
            )
        ).get_success_or_raise()
        return self._validate_page_response(response).items



import hashlib
import json
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Annotated, Any, Literal

from pydantic import ConfigDict, Field, field_validator, model_validator

from ._base import DataSelector, SelectorObject
from ._instances import SelectedView

FILENAME_VARIABLE = "$FILENAME"
FILEPATH = "$FILEPATH"


class FileContentSelector(DataSelector, ABC):
    kind: Literal["FileContent"] = "FileContent"


class FileTemplateSelector(FileContentSelector, ABC):
    file_directory: Path

    def find_data_files(self, input_dir: Path, manifest_file: Path) -> list[Path]:
        file_dir = input_dir / self.file_directory
        if not file_dir.is_dir():
            return []
        return [file for file in file_dir.iterdir() if file.is_file()]

    @abstractmethod
    def create_instance(self, filepath: Path) -> dict[str, Any]: ...


class FileTemplate(SelectorObject):
    model_config = ConfigDict(extra="allow")

    def create_instance(self, filename: str) -> dict[str, Any]:
        json_str = self.model_dump_json(by_alias=True)
        return json.loads(json_str.replace(FILENAME_VARIABLE, filename))


class FileMetadataTemplate(FileTemplate):
    name: str
    external_id: str

    @field_validator("name", "external_id")
    @classmethod
    def _validate_filename_in_fields(cls, v: str) -> str:
        if FILENAME_VARIABLE not in v:
            raise ValueError(
                f"{FILENAME_VARIABLE!s} must be present in 'name' and 'external_id' fields. "
                f"This allows for dynamic substitution based on the file name."
            )
        return v


class FileMetadataTemplateSelector(FileTemplateSelector):
    type: Literal["fileMetadataTemplate"] = "fileMetadataTemplate"
    template: FileMetadataTemplate

    @property
    def group(self) -> str:
        return "FileMetadata"

    def __str__(self) -> str:
        return "metadata_template"

    def create_instance(self, filepath: Path) -> dict[str, Any]:
        return self.template.create_instance(filepath.name)


class TemplateNodeId(SelectorObject):
    space: str
    external_id: str

    @field_validator("external_id")
    @classmethod
    def _validate_filename_in_fields(cls, v: str) -> str:
        if FILENAME_VARIABLE not in v:
            raise ValueError(
                f"{FILENAME_VARIABLE!s} must be present in 'external_id' field. "
                f"This allows for dynamic substitution based on the file name."
            )
        return v


class FileDataModelingTemplate(FileTemplate):
    instance_id: TemplateNodeId
    # Name is required for FileMetadata but not for CogniteFiles. This is the same default behavior as in CDF.
    name: str = "untitled"

    @model_validator(mode="before")
    def _move_space_external_id(cls, data: dict[str, Any]) -> dict[str, Any]:
        if "space" in data and "externalId" in data:
            data["instanceId"] = {"space": data.pop("space"), "externalId": data.pop("externalId")}
        elif "space" in data and "external_id" in data:
            data["instance_id"] = {"space": data.pop("space"), "external_id": data.pop("external_id")}
        return data


class FileDataModelingTemplateSelector(FileTemplateSelector):
    type: Literal["fileDataModelingTemplate"] = "fileDataModelingTemplate"
    view_id: SelectedView = SelectedView(space="cdf_cdm", external_id="CogniteFile", version="v1")
    template: FileDataModelingTemplate

    @property
    def group(self) -> str:
        return "FileDataModeling"

    def __str__(self) -> str:
        return "data_modeling_template"

    def create_instance(self, filepath: Path) -> dict[str, Any]:
        return self.template.create_instance(filepath.name)


class FileIdentifierDefinition(SelectorObject):
    id_type: str


class FileInternalID(FileIdentifierDefinition):
    id_type: Literal["internalId"] = "internalId"
    internal_id: int = Field(alias="id")

    def __str__(self) -> str:
        return f"internalId_{self.internal_id}"


class FileExternalID(FileIdentifierDefinition):
    id_type: Literal["externalId"] = "externalId"
    external_id: str

    def __str__(self) -> str:
        return f"externalId_{self.external_id}"


class NodeId(SelectorObject):
    space: str
    external_id: str


class FileInstanceID(FileIdentifierDefinition):
    id_type: Literal["instanceId"] = "instanceId"
    instance_id: NodeId

    def __str__(self) -> str:
        return f"instanceId_{self.instance_id.space}_{self.instance_id.external_id}"


FileIdentifier = Annotated[FileInstanceID | FileExternalID | FileInternalID, Field(discriminator="id_type")]


class FileIdentifierSelector(FileContentSelector):
    type: Literal["fileIdentifier"] = "fileIdentifier"
    file_directory: str = "file_content"
    use_metadata_directory: bool = True
    identifiers: tuple[FileIdentifier, ...]

    def __str__(self) -> str:
        hash_ = hashlib.md5(",".join(sorted(str(self.identifiers))).encode()).hexdigest()[:8]
        return f"file_{len(self.identifiers)}_identifiers_{hash_}"



from typing import Any

from cognite_toolkit._cdf_tk.client.resource_classes.instance_api import TypedNodeIdentifier

from .base import BaseModelResource, ToolkitResource


class FeatureToggles(BaseModelResource):
    """Feature toggles for InField location configuration."""

    three_d: bool | None = None
    trends: bool | None = None
    documents: bool | None = None
    workorders: bool | None = None
    notifications: bool | None = None
    media: bool | None = None
    template_checklist_flow: bool | None = None


class AccessManagement(BaseModelResource):
    """Access management configuration."""

    template_admins: list[str] | None = None  # list of CDF group external IDs
    checklist_admins: list[str] | None = None  # list of CDF group external IDs


class DataFilterPath(BaseModelResource):
    """Path reference for data filters."""

    space: str
    external_id: str


class DataFilter(BaseModelResource):
    """Data filter configuration for a resource type."""

    path: DataFilterPath | None = None
    instance_spaces: list[str] | None = None


class DataFilters(BaseModelResource):
    """Data filters configuration."""

    files: DataFilter | None = None
    assets: DataFilter | None = None
    operations: DataFilter | None = None
    timeseries: DataFilter | None = None
    notifications: DataFilter | None = None
    maintenance_orders: DataFilter | None = None


class DataStorage(BaseModelResource):
    """Data storage configuration."""

    root_location: DataFilterPath | None = None
    app_instance_space: str | None = None


class ViewMapping(BaseModelResource):
    """View mapping configuration."""

    space: str
    version: str
    external_id: str


class ViewMappings(BaseModelResource):
    """View mappings configuration."""

    asset: ViewMapping | None = None
    operation: ViewMapping | None = None
    notification: ViewMapping | None = None
    maintenance_order: ViewMapping | None = None
    file: ViewMapping | None = None


class Discipline(BaseModelResource):
    """Discipline configuration."""

    name: str
    external_id: str


class InFieldCDMLocationConfigYAML(ToolkitResource):
    """Properties for InFieldCDMLocationConfig node.

    Fields:
    - space: The space of the InField CDM location configuration
    - external_id: The external ID of the InField CDM location configuration
    - name: The name of the location configuration
    - description: The description of the location configuration
    - feature_toggles: Feature toggles configuration
    - access_management: Access management configuration
    - data_filters: Data filters configuration for various resource types
    - data_storage: Data storage configuration
    - view_mappings: View mappings configuration
    - disciplines: List of disciplines
    - data_exploration_config: Data exploration configuration
    """

    space: str
    external_id: str

    name: str | None = None
    description: str | None = None
    feature_toggles: FeatureToggles | None = None
    access_management: AccessManagement | None = None
    data_filters: DataFilters | None = None
    data_storage: DataStorage | None = None
    view_mappings: ViewMappings | None = None
    disciplines: list[Discipline] | None = None
    data_exploration_config: dict[str, Any] | None = None

    def as_id(self) -> TypedNodeIdentifier:
        return TypedNodeIdentifier(space=self.space, external_id=self.external_id)



from pydantic import Field

from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import TransformationNotificationId

from .base import ToolkitResource


class TransformationNotificationYAML(ToolkitResource):
    destination: str = Field(description="Email address where notifications should be sent.")
    transformation_external_id: str = Field(description="Transformation external ID to subscribe.")

    def as_id(self) -> TransformationNotificationId:
        return TransformationNotificationId(
            transformation_external_id=self.transformation_external_id, destination=self.destination
        )



import collections.abc
from collections.abc import Hashable, Iterable, Sequence, Sized
from pathlib import Path
from typing import Any, final

from cognite.client.data_classes.capabilities import Capability, DataModelInstancesAcl

from cognite_toolkit._cdf_tk.client.resource_classes.apm_config_v1 import (
    APM_CONFIG_SPACE,
    APMConfigRequest,
    APMConfigResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import SpaceReference
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling._instance import InstanceSlimDefinition
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId, NameId
from cognite_toolkit._cdf_tk.client.resource_classes.infield import (
    InFieldCDMLocationConfigRequest,
    InFieldCDMLocationConfigResponse,
    InFieldLocationConfigRequest,
    InFieldLocationConfigResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.instance_api import TypedNodeIdentifier
from cognite_toolkit._cdf_tk.constants import BUILD_FOLDER_ENCODING
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.resource_classes import (
    InFieldCDMLocationConfigYAML,
    InfieldLocationConfigYAML,
    InfieldV1YAML,
)
from cognite_toolkit._cdf_tk.utils import quote_int_value_by_key_in_yaml, safe_read
from cognite_toolkit._cdf_tk.utils.diff_list import diff_list_hashable, diff_list_identifiable, hash_dict

from .auth import GroupAllScopedCRUD
from .classic import AssetCRUD
from .data_organization import DataSetsCRUD
from .datamodel import SpaceCRUD, ViewCRUD
from .group_scoped import GroupResourceScopedCRUD


@final
class InfieldV1CRUD(ResourceCRUD[ExternalId, APMConfigRequest, APMConfigResponse]):
    folder_name = "cdf_applications"
    resource_cls = APMConfigResponse
    resource_write_cls = APMConfigRequest
    kind = "InfieldV1"
    yaml_cls = InfieldV1YAML
    dependencies = frozenset({DataSetsCRUD, AssetCRUD, SpaceCRUD, GroupAllScopedCRUD, GroupResourceScopedCRUD})
    _doc_url = "Instances/operation/applyNodeAndEdges"
    _root_location_filters: tuple[str, ...] = ("general", "assets", "files", "timeseries")
    _group_keys: tuple[str, ...] = ("templateAdmins", "checklistAdmins")

    @property
    def display_name(self) -> str:
        return "infield configs"

    @classmethod
    def get_id(cls, item: APMConfigResponse | APMConfigRequest | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        if not item.external_id:
            raise KeyError("APMConfig must have external_id")
        return ExternalId(external_id=item.external_id)

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return {"externalId": id.external_id}

    @classmethod
    def get_required_capability(
        cls, items: collections.abc.Sequence[APMConfigRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [DataModelInstancesAcl.Action.Read]
            if read_only
            else [DataModelInstancesAcl.Action.Read, DataModelInstancesAcl.Action.Write]
        )

        return DataModelInstancesAcl(actions, DataModelInstancesAcl.Scope.SpaceID([APM_CONFIG_SPACE]))

    def prerequisite_warning(self) -> str | None:
        view_id = APMConfigRequest.VIEW_ID
        views = self.client.data_modeling.views.retrieve((view_id.space, view_id.external_id, view_id.version))
        if len(views) > 0:
            return None
        return (
            f"{self.display_name} requires the {APMConfigRequest.VIEW_ID!r} to be deployed. "
            f"Install the infield options with cdf modules init/add to deploy it."
        )

    def create(self, items: Sequence[APMConfigRequest]) -> list[InstanceSlimDefinition]:
        return self.client.infield.apm_config.create(items)

    def retrieve(self, ids: Sequence[ExternalId]) -> list[APMConfigResponse]:
        return self.client.infield.apm_config.retrieve(
            TypedNodeIdentifier.from_external_ids(ids, space=APM_CONFIG_SPACE)
        )

    def update(self, items: Sequence[APMConfigRequest]) -> list[InstanceSlimDefinition]:
        return self.client.infield.apm_config.create(items)

    def delete(self, ids: Sequence[ExternalId]) -> int:
        deleted = self.client.infield.apm_config.delete(
            TypedNodeIdentifier.from_external_ids(ids, space=APM_CONFIG_SPACE)
        )
        return len(deleted)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[APMConfigResponse]:
        raise NotImplementedError(f"Iteration over {self.display_name} is not supported.")

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if isinstance(app_data_space_id := item.get("appDataSpaceId"), str):
            yield SpaceCRUD, SpaceReference(space=app_data_space_id)
        if isinstance(customer_data_space_id := item.get("customerDataSpaceId"), str):
            yield SpaceCRUD, SpaceReference(space=customer_data_space_id)
        for config in cls._get_root_location_configurations(item) or []:
            if isinstance(asset_external_id := config.get("assetExternalId"), str):
                yield AssetCRUD, ExternalId(external_id=asset_external_id)
            if isinstance(data_set_external_id := config.get("dataSetExternalId"), str):
                yield DataSetsCRUD, ExternalId(external_id=data_set_external_id)
            if isinstance(app_data_instance_space := config.get("appDataInstanceSpace"), str):
                yield SpaceCRUD, SpaceReference(space=app_data_instance_space)
            if isinstance(source_data_instance_space := config.get("sourceDataInstanceSpace"), str):
                yield SpaceCRUD, SpaceReference(space=source_data_instance_space)
            for key in cls._group_keys:
                for group in config.get(key, []):
                    if isinstance(group, str):
                        yield GroupResourceScopedCRUD, NameId(name=group)
            data_filters = config.get("dataFilters")
            if not isinstance(data_filters, dict):
                continue
            for key in cls._root_location_filters:
                filter_ = data_filters.get(key)
                if not isinstance(filter_, dict):
                    continue
                for data_set_external_id in filter_.get("dataSetExternalIds", []):
                    if isinstance(data_set_external_id, str):
                        yield DataSetsCRUD, ExternalId(external_id=data_set_external_id)
                for asset_external_id in filter_.get("assetSubtreeExternalIds", []):
                    if isinstance(asset_external_id, str):
                        yield AssetCRUD, ExternalId(external_id=asset_external_id)
                if app_data_instance_space := filter_.get("appDataInstanceSpace"):
                    if isinstance(app_data_instance_space, str):
                        yield SpaceCRUD, SpaceReference(space=app_data_instance_space)

    def safe_read(self, filepath: Path | str) -> str:
        # The customerDataSpaceVersion is a string, but the user often writes it as an int.
        # YAML will then parse it as an int, for example, `3_0_2` will be parsed as `302`.
        # This is technically a user mistake, as you should quote the version in the YAML file.
        # However, we do not want to put this burden on the user (knowing the intricate workings of YAML),
        # so we fix it here.
        return quote_int_value_by_key_in_yaml(
            safe_read(filepath, encoding=BUILD_FOLDER_ENCODING), key="customerDataSpaceVersion"
        )

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> APMConfigRequest:
        root_location_configurations = self._get_root_location_configurations(resource)
        for config in root_location_configurations or []:
            if not isinstance(config, dict):
                continue
            if ds_external_id := config.pop("dataSetExternalId", None):
                config["dataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
            data_filters = config.get("dataFilters")
            if not isinstance(data_filters, dict):
                continue
            for key in self._root_location_filters:
                filter_ = data_filters.get(key)
                if not isinstance(filter_, dict):
                    continue
                if ds_external_ids := filter_.pop("dataSetExternalIds", None):
                    filter_["dataSetIds"] = self.client.lookup.data_sets.id(ds_external_ids, is_dry_run)
        return APMConfigRequest._load(resource)

    def dump_resource(self, resource: APMConfigResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump(context="toolkit")
        local = local or {}
        if "existingVersion" not in local:
            # Existing version is typically not set when creating nodes, but we get it back
            # when we retrieve the node from the server.
            dumped.pop("existingVersion", None)

        for key in ("space", "instanceType"):
            if key in dumped and key not in local:
                # space and instanceType are required when interacting with the API,
                # but not when defining the resource locally in YAML.
                dumped.pop(key)

        for config in self._get_root_location_configurations(dumped) or []:
            if not isinstance(config, dict):
                continue
            if data_set_id := config.pop("dataSetId", None):
                config["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)
            data_filters = config.get("dataFilters")
            if not isinstance(data_filters, dict):
                continue
            for key in self._root_location_filters:
                filter_ = data_filters.get(key)
                if not isinstance(filter_, dict):
                    continue
                if data_set_ids := filter_.pop("dataSetIds", None):
                    filter_["dataSetExternalIds"] = self.client.lookup.data_sets.external_id(data_set_ids)
        return dumped

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path == ("featureConfiguration", "rootLocationConfigurations"):
            return diff_list_identifiable(local, cdf, get_identifier=hash_dict)
        if not (len(json_path) >= 3 and json_path[:2] == ("featureConfiguration", "rootLocationConfigurations")):
            return super().diff_list(local, cdf, json_path)

        if len(json_path) == 4 and json_path[-1] in self._group_keys:
            return diff_list_hashable(local, cdf)
        if len(json_path) == 5 and json_path[-1] in ("fullWeightModels", "lightWeightModels"):
            return diff_list_identifiable(local, cdf, get_identifier=hash_dict)
        if len(json_path) == 6 and json_path[-2] == "dataFilters" and json_path[-1] in self._root_location_filters:
            return diff_list_hashable(local, cdf)
        if len(json_path) == 7 and "observations" in json_path and json_path[-1] in ("type", "priority"):
            return diff_list_identifiable(local, cdf, get_identifier=hash_dict)
        return super().diff_list(local, cdf, json_path)

    @staticmethod
    def _get_root_location_configurations(resource: dict[str, Any]) -> list | None:
        feature_configuration = resource.get("featureConfiguration")
        if not isinstance(feature_configuration, dict):
            return None
        return feature_configuration.get("rootLocationConfigurations")


@final
class InFieldLocationConfigCRUD(
    ResourceCRUD[TypedNodeIdentifier, InFieldLocationConfigRequest, InFieldLocationConfigResponse]
):
    folder_name = "cdf_applications"
    resource_cls = InFieldLocationConfigResponse
    resource_write_cls = InFieldLocationConfigRequest
    kind = "InFieldLocationConfig"
    yaml_cls = InfieldLocationConfigYAML
    dependencies = frozenset({SpaceCRUD, GroupAllScopedCRUD, GroupResourceScopedCRUD})
    _doc_url = "Instances/operation/applyNodeAndEdges"

    @property
    def display_name(self) -> str:
        return "infield location configs"

    @classmethod
    def get_id(cls, item: InFieldLocationConfigRequest | InFieldLocationConfigResponse | dict) -> TypedNodeIdentifier:
        if isinstance(item, dict):
            return TypedNodeIdentifier(space=item["space"], external_id=item["externalId"])
        return TypedNodeIdentifier(space=item.space, external_id=item.external_id)

    @classmethod
    def dump_id(cls, id: TypedNodeIdentifier) -> dict[str, Any]:
        return {
            "space": id.space,
            "externalId": id.external_id,
        }

    @classmethod
    def get_required_capability(
        cls, items: Sequence[InFieldLocationConfigRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items or items is None:
            return []

        actions = (
            [DataModelInstancesAcl.Action.Read]
            if read_only
            else [DataModelInstancesAcl.Action.Read, DataModelInstancesAcl.Action.Write]
        )
        instance_spaces = sorted({item.space for item in items})

        return DataModelInstancesAcl(actions, DataModelInstancesAcl.Scope.SpaceID(instance_spaces))

    def dump_resource(
        self, resource: InFieldLocationConfigResponse, local: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump()
        local = local or {}
        dumped.pop("instanceType", None)
        if isinstance(cdf_dec := dumped.get("dataExplorationConfig"), dict):
            cdf_dec.pop("instanceType", None)
            if isinstance(local_dec := local.get("dataExplorationConfig"), dict):
                if "space" in cdf_dec and "space" not in local_dec:
                    # Default space is used for the data exploration config if not specified locally.
                    cdf_dec.pop("space")
                if "externalId" in cdf_dec and "externalId" not in local_dec:
                    # Default externalId is used for the data exploration config if not specified locally.
                    cdf_dec.pop("externalId")

        return dumped

    def create(self, items: Sequence[InFieldLocationConfigRequest]) -> list[InstanceSlimDefinition]:
        return self.client.infield.config.create(items)

    def retrieve(self, ids: Sequence[TypedNodeIdentifier]) -> list[InFieldLocationConfigResponse]:
        return self.client.infield.config.retrieve(list(ids))

    def update(self, items: Sequence[InFieldLocationConfigRequest]) -> Sized:
        return self.client.infield.config.update(items)

    def delete(self, ids: Sequence[TypedNodeIdentifier]) -> int:
        return len(self.client.infield.config.delete(list(ids)))

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[InFieldLocationConfigResponse]:
        raise NotImplementedError(f"Iteration over {self.display_name} is not supported.")

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path == ("accessManagement", "templateAdmins"):
            return diff_list_hashable(local, cdf)
        elif json_path == ("accessManagement", "checklistAdmins"):
            return diff_list_hashable(local, cdf)
        elif json_path == ("dataFilters", "general", "spaces"):
            return diff_list_hashable(local, cdf)
        elif json_path == ("dataExplorationConfig", "documents", "supportedFormats"):
            return diff_list_hashable(local, cdf)
        return super().diff_list(local, cdf, json_path)


@final
class InFieldCDMLocationConfigCRUD(
    ResourceCRUD[TypedNodeIdentifier, InFieldCDMLocationConfigRequest, InFieldCDMLocationConfigResponse]
):
    folder_name = "cdf_applications"
    resource_cls = InFieldCDMLocationConfigResponse
    resource_write_cls = InFieldCDMLocationConfigRequest
    kind = "InFieldCDMLocationConfig"
    yaml_cls = InFieldCDMLocationConfigYAML
    dependencies = frozenset({SpaceCRUD, GroupAllScopedCRUD, GroupResourceScopedCRUD, ViewCRUD})
    _doc_url = "Instances/operation/applyNodeAndEdges"

    @property
    def display_name(self) -> str:
        return "infield CDM location configs"

    @classmethod
    def get_id(
        cls, item: InFieldCDMLocationConfigRequest | InFieldCDMLocationConfigResponse | dict
    ) -> TypedNodeIdentifier:
        if isinstance(item, dict):
            return TypedNodeIdentifier(space=item["space"], external_id=item["externalId"])
        return TypedNodeIdentifier(space=item.space, external_id=item.external_id)

    @classmethod
    def dump_id(cls, id: TypedNodeIdentifier) -> dict[str, Any]:
        return {
            "space": id.space,
            "externalId": id.external_id,
        }

    @classmethod
    def get_required_capability(
        cls, items: Sequence[InFieldCDMLocationConfigRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items or items is None:
            return []

        actions = (
            [DataModelInstancesAcl.Action.Read]
            if read_only
            else [DataModelInstancesAcl.Action.Read, DataModelInstancesAcl.Action.Write]
        )
        instance_spaces = sorted({item.space for item in items})

        return DataModelInstancesAcl(actions, DataModelInstancesAcl.Scope.SpaceID(instance_spaces))

    def dump_resource(
        self, resource: InFieldCDMLocationConfigResponse, local: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump(context="toolkit")
        local = local or {}
        if "existingVersion" not in local:
            # Existing version is typically not set when creating nodes, but we get it back
            # when we retrieve the node from the server.
            dumped.pop("existingVersion", None)
        dumped.pop("instanceType", None)
        return dumped

    def create(self, items: Sequence[InFieldCDMLocationConfigRequest]) -> list[InstanceSlimDefinition]:
        return self.client.infield.cdm_config.create(items)

    def retrieve(self, ids: Sequence[TypedNodeIdentifier]) -> list[InFieldCDMLocationConfigResponse]:
        return self.client.infield.cdm_config.retrieve(list(ids))

    def update(self, items: Sequence[InFieldCDMLocationConfigRequest]) -> Sized:
        return self.create(items)

    def delete(self, ids: Sequence[TypedNodeIdentifier]) -> int:
        deleted = self.client.infield.cdm_config.delete(list(ids))
        return len(deleted)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[InFieldCDMLocationConfigResponse]:
        raise NotImplementedError(f"Iteration over {self.display_name} is not supported.")

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path == ("accessManagement", "templateAdmins"):
            return diff_list_hashable(local, cdf)
        elif json_path == ("accessManagement", "checklistAdmins"):
            return diff_list_hashable(local, cdf)
        elif json_path == ("disciplines",):
            return diff_list_identifiable(local, cdf, get_identifier=hash_dict)
        elif len(json_path) == 3 and json_path[0] == "dataFilters" and json_path[2] == "instanceSpaces":
            # Handles dataFilters.<entity>.instanceSpaces (e.g., files, assets, operations, timeSeries, etc.)
            return diff_list_hashable(local, cdf)
        elif json_path == ("dataExplorationConfig", "filters"):
            return diff_list_identifiable(local, cdf, get_identifier=hash_dict)
        elif len(json_path) == 4 and json_path[:2] == ("dataExplorationConfig", "filters") and json_path[3] == "values":
            # Handles dataExplorationConfig.filters[i].values
            return diff_list_hashable(local, cdf)

        return super().diff_list(local, cdf, json_path)



from typing import Annotated, Any, Literal

from pydantic import BeforeValidator, Field

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject, RequestResource, ResponseResource

from .identifiers import ExternalId


class AgentToolDefinition(BaseModelObject):
    type: str
    name: str
    description: str


class AnalyzeImage(AgentToolDefinition):
    type: Literal["analyzeImage"] = "analyzeImage"


class AnalyzeTimeSeries(AgentToolDefinition):
    type: Literal["analyzeTimeSeries"] = "analyzeTimeSeries"


class AskDocument(AgentToolDefinition):
    type: Literal["askDocument"] = "askDocument"


class CallFunctionConfig(BaseModelObject):
    external_id: str
    max_polling_time: int = 540
    schema_: dict[str, Any] | None = Field(None, alias="schema")


class CallFunction(AgentToolDefinition):
    type: Literal["callFunction"] = "callFunction"
    configuration: CallFunctionConfig


class CallRestApi(AgentToolDefinition):
    type: Literal["callRestApi"] = "callRestApi"


class ExamineDataSemantically(AgentToolDefinition):
    type: Literal["examineDataSemantically"] = "examineDataSemantically"


class AgentDataModel(BaseModelObject):
    space: str
    external_id: str
    version: str
    view_external_ids: list[str] | None = None


class AgentInstanceSpacesDefinition(BaseModelObject):
    type: str


class AllInstanceSpaces(AgentInstanceSpacesDefinition):
    type: Literal["all"] = "all"


class ManualInstanceSpaces(AgentInstanceSpacesDefinition):
    type: Literal["manual"] = "manual"
    spaces: list[str]


AgentInstanceSpaces = Annotated[
    AllInstanceSpaces | ManualInstanceSpaces,
    Field(discriminator="type"),
]


class QueryKnowledgeGraphConfig(BaseModelObject):
    data_models: list[AgentDataModel]
    instance_spaces: AgentInstanceSpaces | None = None
    # This is deviating from the API documentation, but the Atlas team has confirmed that "v2" is the default
    version: Literal["v1", "v2"] = "v2"


class QueryKnowledgeGraph(AgentToolDefinition):
    type: Literal["queryKnowledgeGraph"] = "queryKnowledgeGraph"
    configuration: QueryKnowledgeGraphConfig


class QueryTimeSeriesDatapoints(AgentToolDefinition):
    type: Literal["queryTimeSeriesDatapoints"] = "queryTimeSeriesDatapoints"


class RunPythonCode(AgentToolDefinition):
    type: Literal["runPythonCode"] = "runPythonCode"


class SummarizeDocument(AgentToolDefinition):
    type: Literal["summarizeDocument"] = "summarizeDocument"


class TimeSeriesAnalysis(AgentToolDefinition):
    type: Literal["timeSeriesAnalysis"] = "timeSeriesAnalysis"


class UnknownAgentTool(AgentToolDefinition):
    """Fallback for unknown tool types."""

    ...


KNOWN_TOOLS: dict[str, type[AgentToolDefinition]] = {
    "analyzeImage": AnalyzeImage,
    "analyzeTimeSeries": AnalyzeTimeSeries,
    "askDocument": AskDocument,
    "callFunction": CallFunction,
    "callRestApi": CallRestApi,
    "examineDataSemantically": ExamineDataSemantically,
    "queryKnowledgeGraph": QueryKnowledgeGraph,
    "queryTimeSeriesDatapoints": QueryTimeSeriesDatapoints,
    "runPythonCode": RunPythonCode,
    "summarizeDocument": SummarizeDocument,
    "timeSeriesAnalysis": TimeSeriesAnalysis,
}


def _handle_unknown_tool(value: Any) -> Any:
    if isinstance(value, dict):
        tool_type = value.get("type")
        if tool_type not in KNOWN_TOOLS:
            return UnknownAgentTool(**value)
        else:
            return KNOWN_TOOLS[tool_type].model_validate(value)
    return value


AgentTool = Annotated[
    AnalyzeImage
    | AnalyzeTimeSeries
    | AskDocument
    | CallFunction
    | CallRestApi
    | ExamineDataSemantically
    | QueryKnowledgeGraph
    | QueryTimeSeriesDatapoints
    | RunPythonCode
    | SummarizeDocument
    | TimeSeriesAnalysis
    | UnknownAgentTool,
    BeforeValidator(_handle_unknown_tool),
]


class Agent(BaseModelObject):
    external_id: str
    name: str
    description: str | None = None
    instructions: str | None = None
    model: str | None = None
    tools: list[AgentTool] | None = None

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)


class AgentRequest(Agent, RequestResource):
    runtime_version: str | None = None


class AgentResponse(Agent, ResponseResource[AgentRequest]):
    created_time: int
    last_updated_time: int
    owner_id: str
    runtime_version: str

    @classmethod
    def request_cls(cls) -> type[AgentRequest]:
        return AgentRequest



"""Temporary here to avoid circular imports while refactoring the codebase."""

from typing import TypeAlias, TypeVar

from cognite.client.data_classes import FileMetadata

from cognite_toolkit._cdf_tk.client.resource_classes.annotation import AnnotationResponse
from cognite_toolkit._cdf_tk.client.resource_classes.asset import AssetResponse
from cognite_toolkit._cdf_tk.client.resource_classes.event import EventResponse
from cognite_toolkit._cdf_tk.client.resource_classes.timeseries import TimeSeriesResponse

AssetCentricResource: TypeAlias = AssetResponse | FileMetadata | EventResponse | TimeSeriesResponse
AssetCentricResourceExtended: TypeAlias = (
    AssetResponse | FileMetadata | EventResponse | TimeSeriesResponse | AnnotationResponse
)
T_AssetCentricResource = TypeVar("T_AssetCentricResource", bound=AssetCentricResource)
T_AssetCentricResourceExtended = TypeVar("T_AssetCentricResourceExtended", bound=AssetCentricResourceExtended)



from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject, ResponseResource, UpdatableRequestResource
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId


class RobotLocation(BaseModelObject):
    """The Locations resource is used to specify the physical location of a robot. Robot missions are defined
    for a specific location. In addition, the location is used to group Missions and Map resources.

    Args:
        name: Location name.
        external_id: Location external id. Must be unique for the resource type.
        description: Description of Location. Textual description of the Location.
    """

    external_id: str
    name: str
    description: str | None = None

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)


class RobotLocationRequest(RobotLocation, UpdatableRequestResource):
    """Request resource for creating or updating a Location."""

    ...


class RobotLocationResponse(RobotLocation, ResponseResource[RobotLocationRequest]):
    """Response resource for a Location."""

    created_time: int
    updated_time: int

    @classmethod
    def request_cls(cls) -> type[RobotLocationRequest]:
        return RobotLocationRequest



from pathlib import Path

from cognite_toolkit._cdf_tk.constants import EXCL_FILES, MODULES
from cognite_toolkit._cdf_tk.cruds import CRUDS_BY_FOLDER_NAME, CRUDS_BY_FOLDER_NAME_INCLUDE_ALPHA
from cognite_toolkit._cdf_tk.data_classes import IssueList
from cognite_toolkit._cdf_tk.data_classes._issues import ModuleLoadingIssue
from cognite_toolkit._cdf_tk.exceptions import ToolkitError
from cognite_toolkit._cdf_tk.utils import humanize_collection, module_path_display_name


class ModulesParser:
    def __init__(self, organization_dir: Path, selected: list[str | Path] | None = None):
        self.organization_dir = organization_dir
        self.selected = selected

    def parse(self) -> tuple[list[Path], IssueList]:
        modules_root = self.organization_dir / MODULES
        if not modules_root.exists():
            raise ToolkitError(f"Module root directory '{modules_root}' not found")

        module_paths: list[Path] = []
        excluded_module_paths: list[Path] = []
        issues: IssueList = IssueList()
        for resource_file in self.organization_dir.glob("**/*.y*ml"):
            if resource_file.name in EXCL_FILES:
                continue

            # Get the module folder for the resource file.
            module_path = self._get_module_path_from_resource_file_path(resource_file)
            if not module_path:
                continue

            # If the module has already been processed, skip it.
            if module_path in module_paths or module_path in excluded_module_paths:
                continue

            # Skip the modules that do not match the selection
            if not self._matches_selection(module_path, modules_root, self.selected):
                excluded_module_paths.append(module_path)
                continue

            module_paths.append(module_path)

        deepest_module_paths = self._find_modules_with_submodules(module_paths)
        parent_module_paths = set(module_paths) - set(deepest_module_paths)
        if parent_module_paths:
            module_paths = deepest_module_paths
            issues.extend(
                ModuleLoadingIssue(
                    message=f"Module {module_path_display_name(self.organization_dir, parent_module_path)!r} is skipped because it has submodules"
                )
                for parent_module_path in parent_module_paths
            )

        valid_module_paths: list[Path] = []
        for module_path in module_paths:
            valid_module_path, issue = self._check_resource_folder_content(module_path)
            if issue:
                issues.append(issue)
            if valid_module_path:
                valid_module_paths.append(valid_module_path)

        return valid_module_paths, issues

    def _get_module_path_from_resource_file_path(self, resource_file: Path) -> Path | None:
        # recognize the module by containing a resource associated by a CRUD.
        # Special case: if the resource folder is a subfolder of a CRUD, return the parent of the subfolder.
        resource_folder = resource_file.parent
        crud = next(iter(CRUDS_BY_FOLDER_NAME_INCLUDE_ALPHA.get(resource_folder.name, [])), None)
        if crud:
            # iterate over the parents of the resource folder until we find the module folder.
            # This is to handle the special case of a subfolder of a CRUD, or yamls in for example function subfolders.
            for p in resource_file.parents:
                if p.name == crud.folder_name:
                    return p.parent
                if p.name == MODULES:
                    return p
        return None

    def _check_resource_folder_content(self, module_path: Path) -> tuple[None | Path, ModuleLoadingIssue | None]:
        resource_folder_names = {d.name for d in module_path.iterdir() if d.is_dir()}
        unrecognized_resource_folder_names = resource_folder_names - CRUDS_BY_FOLDER_NAME.keys()

        issue = (
            ModuleLoadingIssue(
                message=f"Module {module_path_display_name(self.organization_dir, module_path)!r} contains unrecognized resource folder(s): {humanize_collection(unrecognized_resource_folder_names)}"
            )
            if unrecognized_resource_folder_names
            else None
        )

        has_valid_resource_folders = bool(resource_folder_names & CRUDS_BY_FOLDER_NAME.keys())
        return (module_path if has_valid_resource_folders else None, issue)

    def _matches_selection(self, module_path: Path, modules_root: Path, selected: list[str | Path] | None) -> bool:
        if not selected:
            return True

        rel = module_path.relative_to(modules_root)
        rel_parts = [p.lower() for p in rel.parts]
        if not rel_parts:
            # module_path is the modules_root itself
            return False
        name_lower = rel_parts[-1]
        modules_lower = MODULES.lower()

        for sel in selected:
            sel_path = Path(sel) if isinstance(sel, str) else sel

            sel_parts = [p.lower() for p in sel_path.parts]
            if not sel_parts:
                continue

            if sel_parts[0] == modules_lower:
                sel_parts = sel_parts[1:]

            if not sel_parts:
                return True

            if len(sel_parts) == 1 and name_lower == sel_parts[0]:
                return True

            if rel_parts[: len(sel_parts)] == sel_parts:
                return True

        return False

    def _find_modules_with_submodules(self, module_paths: list[Path]) -> list[Path]:
        """Remove parent modules when they have submodules. Keep only the deepest modules."""
        return [
            module_path
            for module_path in module_paths
            if not any(
                module_path in other_module_path.parents
                for other_module_path in module_paths
                if other_module_path != module_path
            )
        ]



from typing import Any, ClassVar, Literal

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    ResponseResource,
    UpdatableRequestResource,
)
from cognite_toolkit._cdf_tk.client._types import Metadata

from .identifiers import ExternalId


class RawTable(BaseModelObject):
    db_name: str
    table_name: str


class Contact(BaseModelObject):
    name: str | None = None
    email: str | None = None
    role: str | None = None
    send_notification: bool | None = None


class NotificationConfig(BaseModelObject):
    allowed_not_seen_range_in_minutes: int | None = None


class ExtractionPipeline(BaseModelObject):
    external_id: str
    name: str
    description: str | None = None
    data_set_id: int
    raw_tables: list[RawTable] | None = None
    schedule: str | None = None
    contacts: list[Contact] | None = None
    metadata: Metadata | None = None
    source: str | None = None
    documentation: str | None = None
    notification_config: NotificationConfig | None = None
    created_by: str | None = None

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)


class ExtractionPipelineRequest(ExtractionPipeline, UpdatableRequestResource):
    container_fields: ClassVar[frozenset[str]] = frozenset({"raw_tables", "contacts", "metadata"})
    non_nullable_fields: ClassVar[frozenset[str]] = frozenset(
        {"documentation", "source", "notification_config", "schedule", "description"}
    )

    def as_update(self, mode: Literal["patch", "replace"]) -> dict[str, Any]:
        update = super().as_update(mode)
        # createdBy cannot be updated.
        update["update"].pop("createdBy", None)
        return update


class ExtractionPipelineResponse(ExtractionPipeline, ResponseResource[ExtractionPipelineRequest]):
    id: int
    created_time: int
    last_updated_time: int

    @classmethod
    def request_cls(cls) -> type[ExtractionPipelineRequest]:
        return ExtractionPipelineRequest



from cognite_toolkit._cdf_tk.client.api.robotics_capabilities import CapabilitiesAPI
from cognite_toolkit._cdf_tk.client.api.robotics_data_postprocessing import DataPostProcessingAPI
from cognite_toolkit._cdf_tk.client.api.robotics_frames import FramesAPI
from cognite_toolkit._cdf_tk.client.api.robotics_locations import LocationsAPI
from cognite_toolkit._cdf_tk.client.api.robotics_maps import MapsAPI
from cognite_toolkit._cdf_tk.client.api.robotics_robots import RobotsAPI
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient


class RoboticsAPI:
    """API for Robotics resources."""

    def __init__(self, http_client: HTTPClient) -> None:
        self.capabilities = CapabilitiesAPI(http_client)
        self.data_postprocessing = DataPostProcessingAPI(http_client)
        self.frames = FramesAPI(http_client)
        self.locations = LocationsAPI(http_client)
        self.maps = MapsAPI(http_client)
        self.robots = RobotsAPI(http_client)



import shutil
import tempfile
import traceback
from collections.abc import Callable
from dataclasses import dataclass
from enum import Enum
from pathlib import Path

import questionary
import typer
from questionary import Choice
from rich import print
from rich.panel import Panel

from cognite_toolkit._cdf_tk.cdf_toml import CDFToml
from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands._base import ToolkitCommand
from cognite_toolkit._cdf_tk.commands.auth import AuthCommand
from cognite_toolkit._cdf_tk.commands.collect import CollectCommand
from cognite_toolkit._cdf_tk.commands.modules import ModulesCommand
from cognite_toolkit._cdf_tk.commands.repo import RepoCommand
from cognite_toolkit._cdf_tk.exceptions import ToolkitError
from cognite_toolkit._cdf_tk.feature_flags import FeatureFlag


class InitItemStatus(Enum):
    """Status of an init checklist item"""

    NONE = "none"
    SUCCESSFUL = "successful"
    FAILED = "failed"


@dataclass
class InitChecklistItem:
    name: str
    description: str
    function: Callable[[], None]
    status: InitItemStatus | None = None
    mandatory: bool = False

    def get_status_display(self) -> str:
        if self.status == InitItemStatus.SUCCESSFUL:
            return "✓"
        elif self.status == InitItemStatus.FAILED:
            return "✗"
        else:
            return "○"

    def get_choice_title(self) -> str:
        status_icon = self.get_status_display()
        return f"{status_icon} {self.description} (required)" if self.mandatory else f"{status_icon} {self.description}"


class InitCommand(ToolkitCommand):
    organization_dir: Path | None

    def __init__(
        self,
        print_warning: bool = True,
        skip_tracking: bool = False,
        silent: bool = False,
        client: ToolkitClient | None = None,
    ) -> None:
        super().__init__(print_warning, skip_tracking, silent, client)
        self.organization_dir = None

    def execute(self, dry_run: bool = False) -> None:
        print("\n")
        print(
            Panel(
                "Go through the following steps to set up the Cognite Toolkit. You can re-run steps if you need to change something.",
                title="Cognite Toolkit Setup",
                style="green",
                padding=(1, 2),
            )
        )

        # Initialize checklist items
        checklist_items = [
            InitChecklistItem(
                name="initToml",
                description="Create toml file",
                function=lambda: self._init_toml(dry_run=dry_run),
                mandatory=True,
            ),
            InitChecklistItem(
                name="initAuth",
                description="Authentication",
                function=lambda: self._init_auth(dry_run=dry_run),
            ),
            InitChecklistItem(
                name="initModules",
                description="Modules",
                function=lambda: self._init_modules(dry_run=dry_run),
            ),
            InitChecklistItem(
                name="initRepo",
                description="Git repository",
                function=lambda: self._init_repo(dry_run=dry_run),
            ),
            InitChecklistItem(
                name="initDataCollection",
                description="Usage statistics",
                function=lambda: self._init_data_collection(dry_run=dry_run),
            ),
        ]

        if CDFToml.load().is_loaded_from_file:
            checklist_items[0].status = InitItemStatus.SUCCESSFUL
            print("cdf.toml configuration file already exists. Skipping creation.")

        # Main loop: keep showing checklist until user is done
        while True:
            # Build choices for questionary
            choices = []
            for item in checklist_items:
                choices.append(
                    Choice(
                        title=item.get_choice_title(),
                        value=item.name,
                    )
                )
            # Check if all mandatory items are completed
            mandatory_items = [item for item in checklist_items if item.mandatory]
            all_mandatory_complete = all(item.status == InitItemStatus.SUCCESSFUL for item in mandatory_items)

            choices.append(Choice(title="> Quit", value="__exit__"))

            # Find the first item with status None to use as default
            default_item = next((item for item in checklist_items if item.status is None), None)
            default_value = default_item.name if default_item else "__exit__"

            # Show checklist and get user selection
            selected = questionary.select(
                "Select a task:",
                choices=choices,
                default=default_value,
            ).unsafe_ask()

            # User cancelled (Ctrl+C or similar)
            if selected is None:
                return

            if selected == "__exit__":
                if all_mandatory_complete:
                    print("Setup complete!")
                    break
                else:
                    incomplete_mandatory = [
                        item.description for item in mandatory_items if item.status != InitItemStatus.SUCCESSFUL
                    ]
                    print(f"Warning: recommended item not completed: {', '.join(incomplete_mandatory)}")
                    break

            # Find the selected item
            selected_item = next((item for item in checklist_items if item.name == selected), None)
            if selected_item is None:
                continue

            # If item was already run, ask for confirmation to re-run
            if selected_item.status is not None:
                status_text = (
                    "successfully. Re-run it?"
                    if selected_item.status == InitItemStatus.SUCCESSFUL
                    else "with failure. Retry?"
                )
                confirm = questionary.confirm(
                    f"'{selected_item.description}' was already run {status_text}",
                    default=False,
                ).unsafe_ask()
                if not confirm:
                    continue

            try:
                selected_item.function()
                selected_item.status = InitItemStatus.SUCCESSFUL
                print(f"✓ {selected_item.description} completed successfully")
            except typer.Exit:
                # typer.Exit is used for normal exits, treat as success
                selected_item.status = InitItemStatus.SUCCESSFUL
                print(f"✓ {selected_item.description} completed successfully")
            except ToolkitError as e:
                # Catch expected toolkit errors
                selected_item.status = InitItemStatus.FAILED
                print(f"✗ {selected_item.description} failed: {e}")
            except Exception as e:
                # Catch unexpected errors and log full traceback for debugging
                selected_item.status = InitItemStatus.FAILED
                print(f"✗ {selected_item.description} failed: {e}")
                print(f"Unexpected error occurred. Full traceback:\n{traceback.format_exc()}")

    def _init_toml(self, dry_run: bool = False) -> None:
        if self.organization_dir is None:
            self.organization_dir = ModulesCommand._prompt_organization_dir()
        if dry_run:
            print("Would initialize cdf.toml configuration file")
            return
        CDFToml.write(self.organization_dir, "dev")
        FeatureFlag.flush()
        print(f"cdf.toml configuration file initialized in {self.organization_dir}")

    def _init_auth(self, dry_run: bool = False) -> None:
        auth_command = AuthCommand()
        auth_command.run(lambda: auth_command.init(no_verify=True, dry_run=dry_run))

    def _init_modules(self, dry_run: bool = False) -> None:
        with ModulesCommand() as modules_command:
            if self.organization_dir is None:
                self.organization_dir = ModulesCommand._prompt_organization_dir()
            if dry_run:
                organization_dir = Path(tempfile.mkdtemp(prefix="init_modules_", suffix=".tmp", dir=Path.cwd()))
                modules_command.run(lambda: modules_command.init(organization_dir=organization_dir))
                shutil.rmtree(organization_dir)
            else:
                modules_command.run(lambda: modules_command.init(organization_dir=self.organization_dir))

    def _init_repo(self, dry_run: bool = False) -> None:
        repo_command = RepoCommand()
        repo_command.run(lambda: repo_command.init(cwd=Path.cwd(), host=None, verbose=False))

    def _init_data_collection(self, dry_run: bool = False) -> None:
        """Opt in to collect usage statistics"""

        opt_in = questionary.confirm(
            "Do you want to opt in to collect usage statistics? This will help us improve the Toolkit.",
            default=True,
        ).unsafe_ask()
        if dry_run:
            print("Would opt in to collect data" if opt_in else "Would not opt in to collect data")
            return

        if opt_in:
            collect_command = CollectCommand()
            collect_command.run(lambda: collect_command.execute("opt-in"))
        else:
            collect_command = CollectCommand()
            collect_command.run(lambda: collect_command.execute("opt-out"))



from pydantic import Field

from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import RawDatabaseId, RawTableId

from .base import ToolkitResource


class DatabaseYAML(ToolkitResource):
    db_name: str = Field(
        description="The name of the database.",
        min_length=1,
        max_length=32,
    )

    def as_id(self) -> RawDatabaseId:
        return RawDatabaseId(name=self.db_name)


class TableYAML(ToolkitResource, populate_by_name=True):
    db_name: str = Field(
        description="The name of the database.",
        min_length=1,
        max_length=32,
    )
    table_name: str = Field(
        description="The name of the table.",
        min_length=1,
        max_length=64,
    )

    def as_id(self) -> RawTableId:
        return RawTableId(db_name=self.db_name, name=self.table_name)



from typing import Annotated, ClassVar, Literal

from pydantic import Field

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    ResponseResource,
    UpdatableRequestResource,
)

from .identifiers import ExternalId


class Mapping(BaseModelObject):
    expression: str


class MappingInputDefinition(BaseModelObject):
    type: str


class ProtobufFile(BaseModelObject):
    file_name: str
    content: str


class ProtobufInput(BaseModelObject):
    type: Literal["protobuf"] = "protobuf"
    message_name: str
    files: list[ProtobufFile]


class CSVInput(BaseModelObject):
    type: Literal["csv"] = "csv"
    delimiter: str = ","
    custom_keys: list[str] | None = None


class XMLInput(BaseModelObject):
    type: Literal["xml"] = "xml"


class JSONInput(BaseModelObject):
    type: Literal["json"] = "json"


MappingInput = Annotated[
    ProtobufInput | CSVInput | XMLInput | JSONInput,
    Field(discriminator="type"),
]


class HostedExtractorMapping(BaseModelObject):
    external_id: str
    mapping: Mapping
    published: bool

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)


class HostedExtractorMappingRequest(HostedExtractorMapping, UpdatableRequestResource):
    non_nullable_fields: ClassVar[frozenset[str]] = frozenset({"input"})
    input: MappingInput | None = None


class HostedExtractorMappingResponse(HostedExtractorMapping, ResponseResource[HostedExtractorMappingRequest]):
    input: MappingInput
    created_time: int
    last_updated_time: int

    @classmethod
    def request_cls(cls) -> type[HostedExtractorMappingRequest]:
        return HostedExtractorMappingRequest



from typing import Annotated, Any, ClassVar, Literal

from pydantic import Field, JsonValue, field_validator

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    ResponseResource,
    UpdatableRequestResource,
)

from .identifiers import ExternalId


class NonceCredentials(BaseModelObject):
    session_id: int
    nonce: str
    cdf_project_name: str
    client_id: str | None = None


class DestinationDefinition(BaseModelObject):
    type: str


class DataModelInfo(BaseModelObject):
    space: str
    external_id: str
    version: str
    destination_type: str
    destination_relationship_from_type: str | None = None


class ViewInfo(BaseModelObject):
    space: str
    external_id: str
    version: str


class EdgeType(BaseModelObject):
    space: str
    external_id: str


class AssetCentricDataSource(DestinationDefinition):
    type: Literal[
        "assets",
        "events",
        "asset_hierarchy",
        "datapoints",
        "string_datapoints",
        "timeseries",
        "sequences",
        "files",
        "labels",
        "relationships",
        "data_sets",
    ]


class DataModelSource(DestinationDefinition):
    type: Literal["instances"] = "instances"
    data_model: DataModelInfo
    instance_space: str | None = None


class ViewDataSource(DestinationDefinition):
    type: Literal["nodes", "edges"]
    view: ViewInfo
    edge_type: EdgeType | None = None
    instance_space: str | None = None


class RawDataSource(DestinationDefinition):
    type: Literal["raw"] = "raw"
    database: str
    table: str


class SequenceRowDataSource(DestinationDefinition):
    type: Literal["sequence_rows"] = "sequence_rows"
    external_id: str


Destination = Annotated[
    AssetCentricDataSource | DataModelSource | ViewDataSource | RawDataSource | SequenceRowDataSource,
    Field(discriminator="type"),
]


class BlockedInfo(BaseModelObject):
    reason: str
    created_time: int


class SessionInfo(BaseModelObject):
    client_id: str | None = None
    session_id: int
    project_name: str


class Transformation(BaseModelObject):
    external_id: str
    name: str
    ignore_null_fields: bool
    data_set_id: int | None = None
    tags: list[str] | None = None

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)


class TransformationRequest(Transformation, UpdatableRequestResource):
    container_fields: ClassVar[frozenset[str]] = frozenset({"tags"})
    non_nullable_fields: ClassVar[frozenset[str]] = frozenset({"is_public", "query", "destination", "conflict_mode"})
    query: str | None = None
    conflict_mode: str | None = None
    destination: Destination | None = None
    source_nonce: NonceCredentials | None = None
    destination_nonce: NonceCredentials | None = None
    is_public: bool | None = None


class TransformationResponse(Transformation, ResponseResource[TransformationRequest]):
    id: int
    created_time: int
    last_updated_time: int
    query: str
    is_public: bool
    conflict_mode: str
    destination: Destination
    blocked: BlockedInfo | None = None
    # In the API spec, owner is documented as a string, but the response is actually {"user": '<uuid>'}
    owner: str | dict[str, str]
    owner_is_current_user: bool
    has_source_oidc_credentials: bool
    has_destination_oidc_credentials: bool
    source_session: SessionInfo | None = None
    destination_session: SessionInfo | None = None
    last_finished_job: dict[str, JsonValue] | None = None
    running_job: dict[str, JsonValue] | None = None
    schedule: dict[str, JsonValue] | None = None

    @classmethod
    def request_cls(cls) -> type[TransformationRequest]:
        return TransformationRequest


class ColumnType(BaseModelObject):
    type: str


class Column(BaseModelObject):
    name: str
    sql_type: str
    type: str | ColumnType
    nullable: bool


class SQLQueryResponse(BaseModelObject):
    schema_: list[Column] = Field(..., alias="schema")
    results: list[dict[str, str | int | float | bool | None]]

    @field_validator("schema_", "results", mode="before")
    def _remove_items_wrapper(cls, value: Any) -> Any:
        if isinstance(value, dict) and "items" in value:
            return value["items"]
        return value



from ._constraints import (
    Constraint,
    ConstraintAdapter,
    ConstraintDefinition,
    RequiresConstraintDefinition,
    UniquenessConstraintDefinition,
)
from ._container import (
    Container,
    ContainerPropertyDefinition,
    ContainerRequest,
    ContainerResponse,
)
from ._data_model import DataModelRequest, DataModelResponse, DataModelResponseWithViews
from ._data_types import (
    BooleanProperty,
    DataType,
    DataTypeAdapter,
    DateProperty,
    DirectNodeRelation,
    EnumProperty,
    EnumValue,
    FileCDFExternalIdReference,
    Float32Property,
    Float64Property,
    FloatProperty,
    Int32Property,
    Int64Property,
    JSONProperty,
    ListablePropertyTypeDefinition,
    PropertyTypeDefinition,
    SequenceCDFExternalIdReference,
    TextProperty,
    TimeseriesCDFExternalIdReference,
    TimestampProperty,
    Unit,
)
from ._indexes import (
    BtreeIndex,
    Index,
    IndexAdapter,
    IndexDefinition,
    InvertedIndex,
)
from ._instance import (
    EdgeRequest,
    EdgeResponse,
    InstanceDefinition,
    InstanceRequest,
    InstanceResponse,
    InstanceResponseDefinition,
    InstanceSource,
    NodeRequest,
    NodeResponse,
)
from ._query import (
    QueryDebugParameters,
    QueryEdgeExpression,
    QueryEdgeTableExpression,
    QueryNodeExpression,
    QueryNodeTableExpression,
    QueryRequest,
    QueryResponse,
    QueryResponseTyped,
    QueryResponseUntyped,
    QuerySelect,
    QuerySelectSource,
    QuerySortSpec,
    QueryTargetUnit,
    QueryThrough,
    QueryUnitReference,
)
from ._references import (
    ContainerConstraintReference,
    ContainerDirectReference,
    ContainerIndexReference,
    ContainerReference,
    DataModelReference,
    DataModelReferenceNoVersion,
    EdgeReference,
    NodeReference,
    SpaceReference,
    ViewDirectReference,
    ViewReference,
    ViewReferenceNoVersion,
)
from ._space import Space, SpaceRequest, SpaceResponse
from ._view import View, ViewRequest, ViewResponse
from ._view_property import (
    ConnectionPropertyDefinition,
    ConstraintOrIndexState,
    EdgeProperty,
    MultiEdgeProperty,
    MultiReverseDirectRelationPropertyRequest,
    MultiReverseDirectRelationPropertyResponse,
    ReverseDirectRelationProperty,
    SingleEdgeProperty,
    SingleReverseDirectRelationPropertyRequest,
    SingleReverseDirectRelationPropertyResponse,
    ViewCoreProperty,
    ViewCorePropertyRequest,
    ViewCorePropertyResponse,
    ViewPropertyDefinition,
    ViewRequestProperty,
    ViewRequestPropertyAdapter,
    ViewResponseProperty,
)

__all__ = [
    "BooleanProperty",
    "BtreeIndex",
    "ConnectionPropertyDefinition",
    "Constraint",
    "ConstraintAdapter",
    "ConstraintDefinition",
    "ConstraintOrIndexState",
    "Container",
    "ContainerConstraintReference",
    "ContainerDirectReference",
    "ContainerIndexReference",
    "ContainerPropertyDefinition",
    "ContainerReference",
    "ContainerRequest",
    "ContainerResponse",
    "DataModelReference",
    "DataModelReferenceNoVersion",
    "DataModelRequest",
    "DataModelResponse",
    "DataModelResponseWithViews",
    "DataType",
    "DataTypeAdapter",
    "DateProperty",
    "DirectNodeRelation",
    "EdgeProperty",
    "EdgeReference",
    "EdgeRequest",
    "EdgeResponse",
    "EnumProperty",
    "EnumValue",
    "FileCDFExternalIdReference",
    "Float32Property",
    "Float64Property",
    "FloatProperty",
    "Index",
    "IndexAdapter",
    "IndexDefinition",
    "InstanceDefinition",
    "InstanceRequest",
    "InstanceResponse",
    "InstanceResponseDefinition",
    "InstanceSource",
    "Int32Property",
    "Int64Property",
    "InvertedIndex",
    "JSONProperty",
    "ListablePropertyTypeDefinition",
    "MultiEdgeProperty",
    "MultiReverseDirectRelationPropertyRequest",
    "MultiReverseDirectRelationPropertyResponse",
    "NodeReference",
    "NodeRequest",
    "NodeResponse",
    "PropertyTypeDefinition",
    "QueryDebugParameters",
    "QueryEdgeExpression",
    "QueryEdgeTableExpression",
    "QueryNodeExpression",
    "QueryNodeTableExpression",
    "QueryRequest",
    "QueryResponse",
    "QueryResponseTyped",
    "QueryResponseUntyped",
    "QuerySelect",
    "QuerySelectSource",
    "QuerySortSpec",
    "QueryTargetUnit",
    "QueryThrough",
    "QueryUnitReference",
    "RequiresConstraintDefinition",
    "ReverseDirectRelationProperty",
    "SequenceCDFExternalIdReference",
    "SingleEdgeProperty",
    "SingleReverseDirectRelationPropertyRequest",
    "SingleReverseDirectRelationPropertyResponse",
    "Space",
    "SpaceReference",
    "SpaceRequest",
    "SpaceResponse",
    "TextProperty",
    "TimeseriesCDFExternalIdReference",
    "TimestampProperty",
    "UniquenessConstraintDefinition",
    "Unit",
    "View",
    "ViewCoreProperty",
    "ViewCorePropertyRequest",
    "ViewCorePropertyResponse",
    "ViewDirectReference",
    "ViewPropertyDefinition",
    "ViewReference",
    "ViewReferenceNoVersion",
    "ViewRequest",
    "ViewRequestProperty",
    "ViewRequestPropertyAdapter",
    "ViewResponse",
    "ViewResponseProperty",
]



import builtins
from typing import ClassVar

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    ResponseResource,
    UpdatableRequestResource,
)
from cognite_toolkit._cdf_tk.client._types import Metadata

from .identifiers import ExternalId, InternalOrExternalId


class Event(BaseModelObject):
    external_id: str | None = None
    data_set_id: int | None = None
    start_time: int | None = None
    end_time: int | None = None
    type: str | None = None
    subtype: str | None = None
    description: str | None = None
    metadata: Metadata | None = None
    asset_ids: list[int] | None = None
    source: str | None = None

    def as_id(self) -> InternalOrExternalId:
        if self.external_id is None:
            raise ValueError("Cannot convert EventRequest to ExternalId when external_id is None")
        return ExternalId(external_id=self.external_id)


class EventRequest(Event, UpdatableRequestResource):
    container_fields: ClassVar[frozenset[str]] = frozenset({"metadata", "asset_ids"})


class EventResponse(Event, ResponseResource[EventRequest]):
    id: int
    created_time: int
    last_updated_time: int

    @classmethod
    def request_cls(cls) -> builtins.type[EventRequest]:
        return EventRequest
